from django.conf import settings
from django.contrib import admin
from django.contrib.auth import get_permission_codename, get_user_model
from django.contrib.auth.models import AnonymousUser
from django.contrib.sites.models import Site
from django.urls import NoReverseMatch, Resolver404, resolve, reverse
from django.utils.translation import (
    gettext_lazy as _,
)
from django.utils.translation import (
    override as force_language,
)

from cms.api import can_change_page
from cms.constants import TEMPLATE_INHERITANCE_MAGIC
from cms.models import Page, PageType, Placeholder
from cms.toolbar.items import REFRESH_PAGE, ButtonList
from cms.toolbar.utils import (
    get_object_edit_url,
    get_object_preview_url,
    get_object_structure_url,
)
from cms.toolbar_base import CMSToolbar
from cms.toolbar_pool import toolbar_pool
from cms.utils import get_language_from_request, page_permissions
from cms.utils.conf import get_cms_setting
from cms.utils.i18n import get_language_dict, get_language_tuple
from cms.utils.page_permissions import (
    user_can_change_page,
    user_can_delete_page,
)
from cms.utils.urlutils import add_url_parameters, admin_reverse
from menus.utils import DefaultLanguageChanger

# Identifiers for search
ADMIN_MENU_IDENTIFIER = 'admin-menu'
"""
The *Site* menu (that usually shows the project's domain name, *example.com* by default).
``ADMIN_MENU_IDENTIFIER`` allows you to get hold of this object easily using
:meth:`cms.toolbar.toolbar.CMSToolbar.get_menu`.
"""
LANGUAGE_MENU_IDENTIFIER = 'language-menu'
"""
The *Language* menu. ``LANGUAGE_MENU_IDENTIFIER`` allows you to get hold of this object
easily using :meth:`cms.toolbar.toolbar.CMSToolbar.get_menu`.
"""
TEMPLATE_MENU_BREAK = 'Template Menu Break'
PAGE_MENU_IDENTIFIER = 'page'
"""
The *Page* menu. ``PAGE_MENU_IDENTIFIER`` allows you to get hold of this object
easily using :meth:`cms.toolbar.toolbar.CMSToolbar.get_menu`.
"""
PAGE_MENU_ADD_IDENTIFIER = 'add_page'
PAGE_MENU_FIRST_BREAK = 'Page Menu First Break'
PAGE_MENU_SECOND_BREAK = 'Page Menu Second Break'
PAGE_MENU_THIRD_BREAK = 'Page Menu Third Break'
PAGE_MENU_FOURTH_BREAK = 'Page Menu Fourth Break'
PAGE_MENU_LAST_BREAK = 'Page Menu Last Break'
HISTORY_MENU_BREAK = 'History Menu Break'
MANAGE_PAGES_BREAK = 'Manage Pages Break'
ADMIN_SITES_BREAK = 'Admin Sites Break'
ADMINISTRATION_BREAK = 'Administration Break'
CLIPBOARD_BREAK = 'Clipboard Break'
USER_SETTINGS_BREAK = 'User Settings Break'
ADD_PAGE_LANGUAGE_BREAK = "Add page language Break"
REMOVE_PAGE_LANGUAGE_BREAK = "Remove page language Break"
COPY_PAGE_LANGUAGE_BREAK = "Copy page language Break"
TOOLBAR_DISABLE_BREAK = 'Toolbar disable Break'
SHORTCUTS_BREAK = 'Shortcuts Break'


@toolbar_pool.register
class PlaceholderToolbar(CMSToolbar):
    """
    Adds placeholder edit buttons if placeholders or static placeholders are detected in the template
    """

    def populate(self):
        self.page = self.request.current_page

    def post_template_populate(self):
        super().post_template_populate()
        self.add_wizard_button()
        self.render_object_editable_buttons()

    def add_wizard_button(self):
        from cms.wizards.wizard_pool import entry_choices
        title = _("Create")

        if self.page:
            user = self.request.user
            page_pk = self.page.pk
            disabled = len(list(entry_choices(user, self.page))) == 0
        else:
            page_pk = ''
            disabled = True

        url = '{url}?page={page}&language={lang}&edit'.format(
            url=reverse("cms_wizard_create"),
            page=page_pk,
            lang=self.toolbar.site_language,
        )
        self.toolbar.add_modal_button(title, url,
                                      side=self.toolbar.RIGHT,
                                      disabled=disabled,
                                      on_close=REFRESH_PAGE)

    def render_object_editable_buttons(self):
        self.init_placeholders()

        if not self.toolbar.obj:
            return

        # Edit button
        if self.toolbar.content_mode_active and self._can_add_button():
            self.add_edit_button()
        # Preview button
        if self.toolbar.edit_mode_active and self._can_add_button():
            self.add_preview_button()
        # Structure mode
        if self._can_add_structure_mode():
            self.add_structure_mode()

    def init_placeholders(self):
        request = self.request
        toolbar = self.toolbar

        if toolbar._async and 'placeholders[]' in request.GET:
            # AJAX request to reload page structure.
            placeholder_ids = request.GET.getlist('placeholders[]')
            self.placeholders = Placeholder.objects.filter(pk__in=placeholder_ids)
        else:
            if toolbar.structure_mode_active and not toolbar.uses_legacy_structure_mode:
                # User has explicitly requested structure mode.
                # and the object (page, blog, etc..) allows for the non-legacy structure mode.
                renderer = toolbar.structure_renderer
            else:
                renderer = toolbar.get_content_renderer()

            self.placeholders = renderer.get_rendered_placeholders()

    # Helpers to check whether buttons can be rendered
    def _has_page_change_perm(self):
        if self.page and user_can_change_page(self.request.user, page=self.page):
            return True
        return False

    def _has_placeholder_change_perm(self):
        if not self.placeholders:
            return False
        return any(
            ph for ph in self.placeholders
            if ph.has_change_permission(self.request.user)
        )

    def _can_add_button(self):
        if self._has_page_change_perm():
            return True
        elif self._has_placeholder_change_perm():
            return True
        return False

    def _can_add_structure_mode(self):
        if not self.request.user.has_perm('cms.use_structure'):
            return False

        if self.page and not self.page.application_urls and self._has_page_change_perm():
            return True
        elif self._has_placeholder_change_perm():
            return True
        return False

    # Buttons
    def add_edit_button(self):
        url = get_object_edit_url(self.toolbar.obj, language=self.toolbar.request_language)
        item = ButtonList(side=self.toolbar.RIGHT)
        item.add_button(
            _('Edit'),
            url=url,
            disabled=False,
            extra_classes=['cms-btn', 'cms-btn-action', 'cms-btn-switch-edit'],
        )
        self.toolbar.add_item(item)

    def add_preview_button(self):
        url = get_object_preview_url(self.toolbar.obj, language=self.toolbar.request_language)
        item = ButtonList(side=self.toolbar.RIGHT)
        item.add_button(
            _('Preview'),
            url=url,
            disabled=False,
            extra_classes=['cms-btn', 'cms-btn-switch-save'],
        )
        self.toolbar.add_item(item)

    def add_structure_mode(self, extra_classes=('cms-toolbar-item-cms-mode-switcher',)):
        structure_active = self.toolbar.structure_mode_active
        edit_mode_active = (not structure_active and self.toolbar.edit_mode_active)
        build_url = get_object_structure_url(self.toolbar.obj, language=self.toolbar.request_language)
        edit_url = get_object_edit_url(self.toolbar.obj, language=self.toolbar.request_language)
        switcher = self.toolbar.add_button_list(
            'Mode Switcher',
            side=self.toolbar.RIGHT,
            extra_classes=extra_classes,
        )
        switcher.add_button(
            _('Structure'),
            build_url,
            active=structure_active,
            disabled=False,
            extra_classes='cms-structure-btn',
        )
        switcher.add_button(
            _('Content'),
            edit_url,
            active=edit_mode_active,
            disabled=False,
            extra_classes='cms-content-btn',
        )


@toolbar_pool.register
class BasicToolbar(CMSToolbar):
    """
    Basic Toolbar for site and languages menu
    """
    page = None
    _language_menu = None
    _admin_menu = None

    def init_from_request(self):
        self.page = self.request.current_page

    def populate(self):
        if not self.page:
            self.init_from_request()
            self.clipboard = self.request.toolbar.user_settings.clipboard
            self.add_admin_menu()
            self.add_language_menu()

    def add_admin_menu(self):
        if not self._admin_menu:
            self._admin_menu = self.toolbar.get_or_create_menu(ADMIN_MENU_IDENTIFIER, self.current_site.name)
            # Users button
            self.add_users_button(self._admin_menu)

            # sites menu
            sites_queryset = Site.objects.order_by('name')

            if len(sites_queryset) > 1:
                sites_menu = self._admin_menu.get_or_create_menu('sites', _('Sites'))
                sites_menu.add_sideframe_item(_('Admin Sites'), url=admin_reverse('sites_site_changelist'))
                sites_menu.add_break(ADMIN_SITES_BREAK)
                for site in sites_queryset:
                    sites_menu.add_link_item(site.name, url='http://%s' % site.domain,
                                             active=site.pk == self.current_site.pk)

            # admin
            self._admin_menu.add_sideframe_item(_('Administration'), url=admin_reverse('index'))
            self._admin_menu.add_break(ADMINISTRATION_BREAK)

            # cms users settings
            self._admin_menu.add_sideframe_item(_('User settings'), url=admin_reverse('cms_usersettings_change'))
            self._admin_menu.add_break(USER_SETTINGS_BREAK)

            # clipboard
            if self.toolbar.edit_mode_active:
                # True if the clipboard exists and there's plugins in it.
                clipboard_is_bound = self.toolbar.clipboard_plugin

                self._admin_menu.add_link_item(
                    _('Clipboard...'), url='#',
                    extra_classes=['cms-clipboard-trigger'],
                    disabled=not clipboard_is_bound
                )
                self._admin_menu.add_link_item(
                    _('Clear clipboard'), url='#',
                    extra_classes=['cms-clipboard-empty'],
                    disabled=not clipboard_is_bound
                )
                self._admin_menu.add_break(CLIPBOARD_BREAK)

            # Disable toolbar
            self._admin_menu.add_link_item(
                _('Disable toolbar'), url='?%s' % get_cms_setting('CMS_TOOLBAR_URL__DISABLE')
            )
            self._admin_menu.add_break(TOOLBAR_DISABLE_BREAK)
            self._admin_menu.add_link_item(
                _('Shortcuts...'), url='#',
                extra_classes=('cms-show-shortcuts',)
            )
            self._admin_menu.add_break(SHORTCUTS_BREAK)

            # logout
            self.add_logout_button(self._admin_menu)

    def add_users_button(self, parent):
        User = get_user_model()

        if User in admin.site._registry:
            opts = User._meta

            if self.request.user.has_perm('%s.%s' % (opts.app_label, get_permission_codename('change', opts))):
                user_changelist_url = admin_reverse('%s_%s_changelist' % (opts.app_label, opts.model_name))
                parent.add_sideframe_item(_('Users'), url=user_changelist_url)

    def add_logout_button(self, parent):
        if self.page and not self.page.login_required:
            anon_can_access = page_permissions.user_can_view_page(
                user=AnonymousUser(),
                page=self.page,
                site=self.current_site,
            )
        else:
            anon_can_access = False

        on_success = self.toolbar.REFRESH_PAGE if anon_can_access else '/'

        # We'll show "Logout Joe Bloggs" if the name fields in auth.User are completed, else "Logout jbloggs". If
        # anything goes wrong, it'll just be "Logout".

        user_name = self.get_username()
        logout_menu_text = _('Logout %s') % user_name if user_name else _('Logout')

        parent.add_ajax_item(
            logout_menu_text,
            action=admin_reverse('logout'),
            active=True,
            on_success=on_success,
            method='GET',
        )

    def add_language_menu(self):
        if settings.USE_I18N and not self._language_menu:
            languages = get_language_tuple(self.current_site.pk)
            if len(languages) > 1:
                # Menu only meaningful if more than one language is installed
                self._language_menu = self.toolbar.get_or_create_menu(
                    LANGUAGE_MENU_IDENTIFIER, _('Language'), position=-1
                )
                language_changer = getattr(self.request, '_language_changer', DefaultLanguageChanger(self.request))
                for code, name in languages:
                    try:
                        url = language_changer(code)
                    except NoReverseMatch:
                        url = DefaultLanguageChanger(self.request)(code)
                    self._language_menu.add_link_item(name, url=url, active=self.current_lang == code)
            else:
                # We do not have to check every time the toolbar is created
                self._language_menu = True  # Pretend the language menu is already there

    def get_username(self, user=None, default=''):
        user = user or self.request.user
        try:
            name = user.get_full_name()
            if name:
                return name
            else:
                return user.get_username()
        except (AttributeError, NotImplementedError):
            return default


@toolbar_pool.register
class PageToolbar(CMSToolbar):
    _changed_admin_menu = None
    watch_models = [Page, PageType]

    def get_page_content(self):
        page_content = self.page.get_content_obj(language=self.current_lang, fallback=False)
        return page_content or None

    def has_page_change_permission(self):
        if not hasattr(self, 'page_change_permission'):
            self.page_change_permission = can_change_page(self.request)
        return self.page_change_permission

    def in_apphook(self):
        with force_language(self.toolbar.request_language):
            try:
                resolver = resolve(self.toolbar.request_path)
            except Resolver404:
                return False
            else:
                cms_views = (
                    "render_object_edit", "render_object_preview", "render_object_structure", "details"
                )
                return resolver.func.__name__ not in cms_views

    def in_apphook_root(self):
        """
        Returns True if the request is for a page handled by an apphook, but
        is also the page it is attached to.
        :return: Boolean
        """
        page = getattr(self.request, 'current_page', False)
        if page:
            language = get_language_from_request(self.request)
            return self.toolbar.request_path == page.get_absolute_url(language=language)
        return False

    def get_on_delete_redirect_url(self):
        language = self.current_lang
        parent_page = self.page.parent_page if self.page else None

        # if the current page has a parent in the request's current language redirect to it
        if parent_page and language in parent_page.get_languages():
            with force_language(language):
                return parent_page.get_absolute_url(language=language)

        # else redirect to root, do not redirect to Page.objects.get_home() because user could have deleted the last
        # page, if DEBUG == False this could cause a 404
        return reverse('pages-root')

    # Populate

    def populate(self):
        self.page = self.request.current_page
        self.title = self.get_page_content() if self.page else None
        self.permissions_activated = get_cms_setting('PERMISSION')
        self.change_admin_menu()
        self.add_page_menu()
        self.change_language_menu()

    # Menus
    def change_language_menu(self):
        if self.toolbar.edit_mode_active and self.page:
            can_change = page_permissions.user_can_change_page(
                user=self.request.user,
                page=self.page,
                site=self.current_site,
            )
        else:
            can_change = False

        if can_change:
            language_menu = self.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
            if not language_menu:
                return None

            languages = get_language_dict(self.current_site.pk)

            remove = [(code, languages.get(code, code)) for code in self.page.get_languages() if code in languages]
            add = [lang for lang in languages.items() if lang not in remove]
            copy = [
                (code, name) for code, name in languages.items() if
                code != self.current_lang and (code, name) in remove
            ]

            if add or remove or copy:
                language_menu.add_break(ADD_PAGE_LANGUAGE_BREAK)

            if add:
                add_plugins_menu = language_menu.get_or_create_menu(
                    '{0}-add'.format(LANGUAGE_MENU_IDENTIFIER), _('Add Translation')
                )

                page_add_url = admin_reverse('cms_pagecontent_add')

                for code, name in add:
                    url = add_url_parameters(page_add_url, cms_page=self.page.pk, language=code)
                    add_plugins_menu.add_modal_item(name, url=url)

            if remove:
                remove_plugins_menu = language_menu.get_or_create_menu(
                    '{0}-del'.format(LANGUAGE_MENU_IDENTIFIER), _('Delete Translation')
                )
                disabled = len(remove) == 1
                for code, name in remove:
                    pagecontent = self.page.get_content_obj(code)
                    translation_delete_url = admin_reverse('cms_pagecontent_delete', args=(pagecontent.pk,))
                    url = add_url_parameters(translation_delete_url, language=code)
                    remove_plugins_menu.add_modal_item(name, url=url, disabled=disabled)

            if copy:
                copy_plugins_menu = language_menu.get_or_create_menu(
                    '{0}-copy'.format(LANGUAGE_MENU_IDENTIFIER), _('Copy all plugins')
                )
                title = _('from %s')
                question = _('Are you sure you want to copy all plugins from %s?')

                for code, name in copy:
                    pagecontent = self.page.get_content_obj(code)
                    page_copy_url = admin_reverse('cms_pagecontent_copy_language', args=(pagecontent.pk,))
                    copy_plugins_menu.add_ajax_item(
                        title % name, action=page_copy_url,
                        data={'source_language': code, 'target_language': self.current_lang},
                        question=question % name, on_success=self.toolbar.REFRESH_PAGE
                    )

    def change_admin_menu(self):
        can_change_page = self.has_page_change_permission()

        if not can_change_page:
            # Check if the user has permissions to change at least one page
            can_change_page = page_permissions.user_can_change_at_least_one_page(
                user=self.request.user,
                site=self.current_site,
            )

        if not self._changed_admin_menu and can_change_page:
            admin_menu = self.toolbar.get_or_create_menu(ADMIN_MENU_IDENTIFIER)
            url = admin_reverse('cms_pagecontent_changelist')  # cms page admin
            params = {'language': self.toolbar.request_language}
            if self.page:
                params['page_id'] = self.page.pk
            url = add_url_parameters(url, params)
            admin_menu.add_sideframe_item(_('Pages'), url=url, position=0)
            # Used to prevent duplicates
            self._changed_admin_menu = True

    def add_page_menu(self):
        if self.page and self.title:
            edit_mode = self.toolbar.edit_mode_active
            refresh = self.toolbar.REFRESH_PAGE
            can_change = user_can_change_page(
                user=self.request.user,
                page=self.page,
                site=self.current_site,
            )

            # menu for current page
            # NOTE: disabled if the current path is "deeper" into the
            # application's url patterns than its root. This is because
            # when the Content Manager is at the root of the app-hook,
            # some of the page options still make sense.
            current_page_menu = self.toolbar.get_or_create_menu(
                PAGE_MENU_IDENTIFIER, _('Page'), position=1, disabled=self.in_apphook() and not self.in_apphook_root())

            new_page_params = {'edit': 1}
            new_sub_page_params = {'edit': 1, 'parent_node': self.page.node_id}

            add_page_url = admin_reverse('cms_pagecontent_add')
            advanced_url = admin_reverse('cms_page_advanced', args=(self.page.pk,))
            page_settings_url = admin_reverse('cms_pagecontent_change', args=(self.title.pk,))
            duplicate_page_url = admin_reverse('cms_pagecontent_duplicate', args=[self.title.pk])

            can_add_root_page = page_permissions.user_can_add_page(
                user=self.request.user,
                site=self.current_site,
            )

            if self.page.parent_page:
                new_page_params['parent_node'] = self.page.parent_page.node_id
                can_add_sibling_page = page_permissions.user_can_add_subpage(
                    user=self.request.user,
                    target=self.page.parent_page,
                )
            else:
                can_add_sibling_page = can_add_root_page

            can_add_sub_page = page_permissions.user_can_add_subpage(
                user=self.request.user,
                target=self.page,
            )

            # page operations menu
            add_page_menu = current_page_menu.get_or_create_menu(
                PAGE_MENU_ADD_IDENTIFIER,
                _('Create Page'),
            )

            add_page_menu_modal_items = (
                (_('New Page'), new_page_params, can_add_sibling_page),
                (_('New Sub Page'), new_sub_page_params, can_add_sub_page),
            )

            for title, params, has_perm in add_page_menu_modal_items:
                params.update(language=self.toolbar.request_language)
                add_page_menu.add_modal_item(
                    title,
                    url=add_url_parameters(add_page_url, params),
                    disabled=not has_perm,
                )

            add_page_menu.add_modal_item(
                _('Duplicate this Page'),
                url=add_url_parameters(duplicate_page_url, {'language': self.toolbar.request_language}),
                disabled=not can_add_sibling_page,
            )

            # first break
            current_page_menu.add_break(PAGE_MENU_FIRST_BREAK)

            # page edit
            with force_language(self.current_lang):
                page_edit_url = get_object_edit_url(self.title) if self.title else ''
                current_page_menu.add_link_item(_('Edit this Page'), disabled=edit_mode, url=page_edit_url)

            # page settings
            page_settings_url = add_url_parameters(page_settings_url, language=self.toolbar.request_language)
            settings_disabled = not edit_mode or not can_change
            current_page_menu.add_modal_item(_('Page settings'), url=page_settings_url, disabled=settings_disabled,
                                             on_close=refresh)

            # advanced settings
            advanced_url = add_url_parameters(advanced_url, language=self.toolbar.request_language)
            can_change_advanced = self.page.has_advanced_settings_permission(self.request.user)
            advanced_disabled = not edit_mode or not can_change_advanced
            current_page_menu.add_modal_item(_('Advanced settings'), url=advanced_url, disabled=advanced_disabled)

            # templates menu
            if edit_mode:
                action = admin_reverse('cms_pagecontent_change_template', args=(self.title.pk,))

                if can_change_advanced:
                    templates_menu = current_page_menu.get_or_create_menu(
                        'templates',
                        _('Templates'),
                        disabled=not can_change,
                    )

                    for path, name in get_cms_setting('TEMPLATES'):
                        active = self.page.template == path
                        if path == TEMPLATE_INHERITANCE_MAGIC:
                            templates_menu.add_break(TEMPLATE_MENU_BREAK)
                        templates_menu.add_ajax_item(name, action=action, data={'template': path}, active=active,
                                                     on_success=refresh)

            # navigation toggle
            in_navigation = self.title.in_navigation
            nav_title = _('Hide in navigation') if in_navigation else _('Display in navigation')
            nav_action = admin_reverse('cms_pagecontent_change_innavigation', args=(self.title.pk,))
            current_page_menu.add_ajax_item(
                nav_title,
                action=nav_action,
                disabled=(not edit_mode or not can_change),
                on_success=refresh,
            )

            # delete
            delete_url = admin_reverse('cms_page_delete', args=(self.page.pk,))
            delete_disabled = not edit_mode or not user_can_delete_page(self.request.user, page=self.page)
            on_delete_redirect_url = self.get_on_delete_redirect_url()
            current_page_menu.add_modal_item(_('Delete page'), url=delete_url, on_close=on_delete_redirect_url,
                                             disabled=delete_disabled)



from cms.cache.permissions import clear_user_permission_cache
from cms.models import PageUser, PageUserGroup
from menus.menu_pool import menu_pool


def post_save_user(instance, raw, created, **kwargs):
    """Signal called when new user is created, required only when CMS_PERMISSION.
    Assigns creator of the user to PageUserInfo model, so we know who had created
    this user account.

    requires: CurrentUserMiddleware
    """
    from cms.utils.permissions import get_current_user

    # read current user from thread locals
    creator = get_current_user()
    if not creator or not created or creator.is_anonymous:
        return

    page_user = PageUser(user_ptr_id=instance.pk, created_by=creator)
    page_user.__dict__.update(instance.__dict__)
    page_user.save()


def post_save_user_group(instance, raw, created, **kwargs):
    """The same like post_save_user, but for Group, required only when
    CMS_PERMISSION.
    Assigns creator of the group to PageUserGroupInfo model, so we know who had
    created this user account.

    requires: CurrentUserMiddleware
    """
    from cms.utils.permissions import get_current_user

    # read current user from thread locals
    creator = get_current_user()
    if not creator or not created or creator.is_anonymous:
        return
    page_user = PageUserGroup(group_ptr_id=instance.pk, created_by=creator)
    page_user.__dict__.update(instance.__dict__)
    page_user.save()


def pre_save_user(instance, raw, **kwargs):
    clear_user_permission_cache(instance)


def pre_delete_user(instance, **kwargs):
    clear_user_permission_cache(instance)


def pre_save_group(instance, raw, **kwargs):
    if instance.pk:
        user_set = getattr(instance, 'user_set')
        for user in user_set.all():
            clear_user_permission_cache(user)


def pre_delete_group(instance, **kwargs):
    user_set = getattr(instance, 'user_set')
    for user in user_set.all():
        clear_user_permission_cache(user)


def _clear_users_permissions(instance):
    if instance.user:
        clear_user_permission_cache(instance.user)
    if instance.group:
        user_set = getattr(instance.group, 'user_set')
        for user in user_set.all():
            clear_user_permission_cache(user)


def pre_save_pagepermission(instance, raw, **kwargs):
    _clear_users_permissions(instance)


def pre_delete_pagepermission(instance, **kwargs):
    _clear_users_permissions(instance)


def pre_save_globalpagepermission(instance, raw, **kwargs):
    _clear_users_permissions(instance)
    menu_pool.clear(all=True)


def pre_delete_globalpagepermission(instance, **kwargs):
    _clear_users_permissions(instance)



from cms.management.commands.subcommands.list import plugin_report

from .base import SubcommandsCommand


class DeleteOrphanedPluginsCommand(SubcommandsCommand):
    help_string = ('Delete plugins from the CMSPlugins table that should have instances '
                   'but don\'t, and ones for which a corresponding plugin model can no '
                   'longer be found')
    command_name = 'delete-orphaned-plugins'

    def handle(self, *args, **options):
        """
        Obtains a plugin report -
        cms.management.commands.subcommands.list.plugin_report - and uses it
        to delete orphaned plugins from the database, i.e. ones that are no
        longer installed, and ones that have no corresponding saved plugin
        instances (as will happen if a plugin is inserted into a placeholder,
        but not saved).
        """
        self.stdout.write('Obtaining plugin report\n')
        uninstalled_instances = []
        unsaved_instances = []

        for plugin in plugin_report():
            if not plugin['model']:
                for instance in plugin['instances']:
                    uninstalled_instances.append(instance)

            for instance in plugin['unsaved_instances']:
                unsaved_instances.append(instance)

        if options.get('interactive'):
            confirm = input("""
You have requested to delete any instances of uninstalled plugins and empty plugin instances.
There are %d uninstalled plugins and %d empty plugins.
Are you sure you want to do this?
Type 'yes' to continue, or 'no' to cancel: """ % (len(uninstalled_instances), len(unsaved_instances)))
        else:
            confirm = 'yes'

        if confirm == 'yes':
            # delete items whose plugin is uninstalled and items with unsaved instances
            self.stdout.write('... deleting any instances of uninstalled plugins and empty plugin instances\n')

            for instance in uninstalled_instances:
                instance.delete()

            for instance in unsaved_instances:
                instance.delete()

            self.stdout.write(
                'Deleted instances of: \n    %s uninstalled plugins  \n    %s plugins with unsaved instances\n' % (
                    len(uninstalled_instances), len(unsaved_instances)
                )
            )
            self.stdout.write('all done\n')



class PluginAlreadyRegistered(Exception):
    pass


class PluginNotRegistered(Exception):
    pass


class PluginLimitReached(Exception):
    """
    Gets triggered when a placeholder has reached it's plugin limit.
    """
    pass


class AppAlreadyRegistered(Exception):
    pass


class ToolbarAlreadyRegistered(Exception):
    pass


class ToolbarNotRegistered(Exception):
    pass


class NotImplemented(Exception):
    pass


class SubClassNeededError(Exception):
    pass


class MissingFormError(Exception):
    pass


class NoHomeFound(Exception):
    pass


class PermissionsException(Exception):
    """Base permission exception
    """


class NoPermissionsException(PermissionsException):
    """Can be fired when some violate action is performed on permission system.
    """


class PublicIsUnmodifiable(Exception):
    """A method was invoked on the public copy, but is only valid for the
    draft version"""
    pass


class PublicVersionNeeded(Exception):
    """A Public version of this page is needed"""
    pass


class Deprecated(Exception): pass


class DuplicatePlaceholderWarning(Warning): pass


class DontUsePageAttributeWarning(Warning): pass


class CMSDeprecationWarning(Warning): pass


class LanguageError(Exception): pass


class PluginConsistencyError(Exception): pass


class PlaceholderNotFound(Exception): pass


class ConfirmationOfVersion4Required(Exception):
    pass



from copy import deepcopy

from django.contrib import admin
from django.contrib.admin import site
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.utils.translation import gettext

from cms.admin.forms import PageUserChangeForm, PageUserGroupForm
from cms.exceptions import NoPermissionsException
from cms.models import Page, PagePermission, PageUser, PageUserGroup
from cms.utils.compat.forms import UserAdmin
from cms.utils.conf import get_cms_setting
from cms.utils.permissions import (
    get_model_permission_codename,
    get_subordinate_groups,
    get_subordinate_users,
    get_user_permission_level,
)

user_model = get_user_model()
admin_class = UserAdmin
for model, admin_instance in site._registry.items():
    if model == user_model:
        admin_class = admin_instance.__class__


class GenericCmsPermissionAdmin():

    def get_subordinates(self, user, site):
        raise NotImplementedError

    def _has_change_permissions_permission(self, request):
        """
        User is able to add/change objects only if he haves can change
        permission on some page.
        """
        site = Site.objects.get_current(request)

        try:
            get_user_permission_level(request.user, site)
        except NoPermissionsException:
            return False
        return True

    def get_form(self, request, obj=None, **kwargs):
        form_class = super().get_form(request, obj, **kwargs)
        form_class._current_user = request.user
        return form_class

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        site = Site.objects.get_current(request)
        user_ids = self.get_subordinates(request.user, site).values_list('pk', flat=True)
        return queryset.filter(pk__in=user_ids)

    def has_add_permission(self, request):
        has_model_perm = super().has_add_permission(request)

        if not has_model_perm:
            return False
        return self._has_change_permissions_permission(request)

    def has_change_permission(self, request, obj=None):
        has_model_perm = super().has_change_permission(request, obj)

        if not has_model_perm:
            return False
        return self._has_change_permissions_permission(request)

    def has_delete_permission(self, request, obj=None):
        has_model_perm = super().has_delete_permission(request, obj)

        if not has_model_perm:
            return False
        return self._has_change_permissions_permission(request)

    def has_view_permission(self, request, obj=None):
        # For django 2.1
        # Default is to return True if user got `change` perm, but we have to
        # get in consideration also cms permission system
        return self.has_change_permission(request, obj)


class PageUserAdmin(GenericCmsPermissionAdmin, admin_class):
    form = PageUserChangeForm
    model = PageUser

    def get_subordinates(self, user, site):
        return get_subordinate_users(user, site).values_list('pk', flat=True)

    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)

        if not request.user.is_superuser:
            # Non superusers can't set superuser status on
            # their subordinates.
            fields = list(fields) + ['is_superuser']
        return fields

    def save_model(self, request, obj, form, change):
        if not change:
            # By default set the staff flag to True
            # when a PageUser is first created
            obj.is_staff = True
            # Set the created_by field to the current user
            obj.created_by = request.user
        super().save_model(request, obj, form, change)


class PageUserGroupAdmin(GenericCmsPermissionAdmin, admin.ModelAdmin):
    form = PageUserGroupForm
    list_display = ('name', 'created_by')

    fieldsets = [
        (None, {'fields': ('name',)}),
    ]

    def get_fieldsets(self, request, obj=None):
        """
        Nobody can grant more than he haves, so check for user permissions
        to Page and User model and render fieldset depending on them.
        """
        fieldsets = deepcopy(self.fieldsets)
        perm_models = (
            (Page, gettext('Page permissions')),
            (PageUser, gettext('User & Group permissions')),
            (PagePermission, gettext('Page permissions management')),
        )
        for i, perm_model in enumerate(perm_models):
            fields = []
            model, title = perm_model
            name = model.__name__.lower()
            for key in ('add', 'change', 'delete'):
                perm_code = get_model_permission_codename(model, action=key)
                if request.user.has_perm(perm_code):
                    fields.append('can_%s_%s' % (key, name))
            if fields:
                fieldsets.insert(2 + i, (title, {'fields': (fields,)}))
        return fieldsets

    def get_subordinates(self, user, site):
        return get_subordinate_groups(user, site).values_list('pk', flat=True)


if get_cms_setting('PERMISSION'):
    admin.site.register(PageUser, PageUserAdmin)
    admin.site.register(PageUserGroup, PageUserGroupAdmin)



import copy

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, Group, Permission
from django.contrib.sites.models import Site
from django.template import Template, TemplateSyntaxError
from django.template.context import Context
from django.test.utils import override_settings
from django.utils.translation import activate
from django.utils.translation import override as force_language

from cms.api import create_page, create_page_content
from cms.apphook_pool import apphook_pool
from cms.cms_menus import get_visible_nodes
from cms.models import ACCESS_PAGE_AND_DESCENDANTS, Page
from cms.models.permissionmodels import GlobalPagePermission, PagePermission
from cms.test_utils.fixtures.menus import (
    ExtendedMenusFixture,
    MenusFixture,
    SoftrootFixture,
    SubMenusFixture,
)
from cms.test_utils.project.sampleapp.cms_apps import (
    NamespacedApp,
    SampleApp,
    SampleApp2,
)
from cms.test_utils.project.sampleapp.cms_menus import (
    SampleAppMenu,
    StaticMenu,
    StaticMenu2,
)
from cms.test_utils.testcases import CMSTestCase
from cms.test_utils.util.context_managers import LanguageOverride, apphooks
from cms.test_utils.util.mock import AttributeObject
from cms.utils import get_current_site
from cms.utils.conf import get_cms_setting
from menus.base import NavigationNode
from menus.menu_pool import _build_nodes_inner_for_one_menu, menu_pool
from menus.models import CacheKey
from menus.utils import cut_levels, find_selected, mark_descendants


class BaseMenuTest(CMSTestCase):

    def _get_nodes(self, path='/'):
        node1 = NavigationNode('1', '/1/', 1)
        node2 = NavigationNode('2', '/2/', 2, 1)
        node3 = NavigationNode('3', '/3/', 3, 2)
        node4 = NavigationNode('4', '/4/', 4, 2)
        node5 = NavigationNode('5', '/5/', 5)
        nodes = [node1, node2, node3, node4, node5]
        tree = _build_nodes_inner_for_one_menu([n for n in nodes], "test")
        request = self.get_request(path)
        renderer = menu_pool.get_renderer(request)
        renderer.apply_modifiers(tree, request)
        return tree, nodes

    def setUp(self):
        super().setUp()
        if not menu_pool.discovered:
            menu_pool.discover_menus()
        self.old_menu = menu_pool.menus
        menu_pool.menus = {'CMSMenu': self.old_menu['CMSMenu']}
        menu_pool.clear(settings.SITE_ID)
        activate("en")

    def tearDown(self):
        menu_pool.menus = self.old_menu
        super().tearDown()

    def get_page(self, num):
        return Page.objects.get(pagecontent_set__title='P%s' % num)


class MenuDiscoveryTest(ExtendedMenusFixture, CMSTestCase):

    def setUp(self):
        super().setUp()
        menu_pool.discovered = False
        self.old_menu = menu_pool.menus
        menu_pool.menus = {}
        menu_pool.discover_menus()
        menu_pool.register_menu(SampleAppMenu)
        menu_pool.register_menu(StaticMenu)
        menu_pool.register_menu(StaticMenu2)

    def tearDown(self):
        menu_pool.menus = self.old_menu
        super().tearDown()

    def test_menu_registered(self):
        menu_pool.discovered = False
        menu_pool.discover_menus()

        # The following tests that get_registered_menus()
        # returns all menus registered based on the for_rendering flag

        # A list of menu classes registered regardless of whether they
        # have instances attached or not
        registered = menu_pool.get_registered_menus(for_rendering=False)

        # A list of menu classes registered and filter out any attached menu
        # if it does not have instances.
        registered_for_rendering = menu_pool.get_registered_menus(for_rendering=True)

        # We've registered three menus
        self.assertEqual(len(registered), 3)

        # But two of those are attached menus and shouldn't be rendered.
        self.assertEqual(len(registered_for_rendering), 1)

        # Attached both menus to separate pages
        create_page("apphooked-page", "nav_playground.html", "en",
                    navigation_extenders='StaticMenu')

        create_page("apphooked-page", "nav_playground.html", "en",
                    navigation_extenders='StaticMenu2')

        registered = menu_pool.get_registered_menus(for_rendering=False)
        registered_for_rendering = menu_pool.get_registered_menus(for_rendering=True)

        # Even though we've registered three menus, the total is give because two
        # are attached menus and each attached menu has two instances.
        self.assertEqual(len(registered), 3)
        self.assertEqual(len(registered_for_rendering), 3)

    def test_menu_registered_in_renderer(self):
        menu_pool.discovered = False
        menu_pool.discover_menus()

        # The following tests that a menu renderer calculates the registered
        # menus on a request basis.

        request_1 = self.get_request('/en/')
        request_1_renderer = menu_pool.get_renderer(request_1)

        registered = menu_pool.get_registered_menus(for_rendering=False)

        self.assertEqual(len(registered), 3)
        self.assertEqual(len(request_1_renderer.menus), 1)

        create_page("apphooked-page", "nav_playground.html", "en",
                    navigation_extenders='StaticMenu')

        create_page("apphooked-page", "nav_playground.html", "en",
                    navigation_extenders='StaticMenu2')

        request_2 = self.get_request('/en/')
        request_2_renderer = menu_pool.get_renderer(request_2)
        self.assertEqual(len(request_2_renderer.menus), 3)

    def test_menu_expanded(self):
        menu_pool.discovered = False
        menu_pool.discover_menus()

        with self.settings(ROOT_URLCONF='cms.test_utils.project.urls_for_apphook_tests'):
            with apphooks(SampleApp):
                page = create_page("apphooked-page", "nav_playground.html", "en",
                                   apphook="SampleApp",
                                   navigation_extenders='StaticMenu')

                self.assertTrue(menu_pool.discovered)

                menus = menu_pool.get_registered_menus()

                self.assertTrue(menu_pool.discovered)
                # Counts the number of StaticMenu (which is expanded) and StaticMenu2
                # (which is not) and checks the key name for the StaticMenu instances
                static_menus = 1
                static_menus_2 = 1
                for key, menu in menus.items():
                    if key.startswith('StaticMenu:'):
                        static_menus -= 1
                        self.assertTrue(key.endswith(str(page.pk)) or key.endswith(str(page.pk)))

                    if key == 'StaticMenu2':
                        static_menus_2 -= 1

                self.assertEqual(static_menus, 0)
                self.assertEqual(static_menus_2, 0)

    def test_multiple_menus(self):
        with self.settings(ROOT_URLCONF='cms.test_utils.project.urls_for_apphook_tests'):
            with apphooks(NamespacedApp, SampleApp2):
                apphook_pool.discovered = False
                apphook_pool.discover_apps()
                create_page("apphooked-page", "nav_playground.html", "en",
                            apphook="SampleApp2")
                create_page("apphooked-page", "nav_playground.html", "en",
                            navigation_extenders='StaticMenu')
                create_page("apphooked-page", "nav_playground.html", "en",
                            apphook="NamespacedApp", apphook_namespace='whatever',
                            navigation_extenders='StaticMenu')
                self.assertEqual(len(menu_pool.get_menus_by_attribute("cms_enabled", True)), 2)


class ExtendedFixturesMenuTests(ExtendedMenusFixture, BaseMenuTest):
    """
    Tree from fixture:

        + P1
        | + P2
        |   + P3
        | + P9
        |   + P10
        |      + P11
        + P4
        | + P5
        + P6 (not in menu)
          + P7
          + P8
    """
    def get_page(self, num):
        return Page.objects.get(pagecontent_set__title='P%s' % num)

    def get_all_pages(self):
        return Page.objects.all()

    def test_menu_failfast_on_invalid_usage(self):
        context = self.get_context()
        context['child'] = self.get_page(1)
        # test standard show_menu
        with self.settings(DEBUG=True, TEMPLATE_DEBUG=True):
            tpl = Template("{% load menu_tags %}{% show_menu 0 0 0 0 'menu/menu.html' child %}")
            self.assertRaises(TemplateSyntaxError, tpl.render, context)

    def test_show_submenu_nephews(self):
        page_2 = self.get_page(2)
        context = self.get_context(path=page_2.get_absolute_url(), page=page_2)
        tpl = Template("{% load menu_tags %}{% show_sub_menu 100 1 1 %}")
        tpl.render(context)
        nodes = context["children"]
        # P2 is the selected node
        self.assertTrue(nodes[0].selected)
        # Should include P10 but not P11
        self.assertEqual(len(nodes[1].children), 1)
        self.assertFalse(nodes[1].children[0].children)

        tpl = Template("{% load menu_tags %}{% show_sub_menu 100 1 %}")
        tpl.render(context)
        nodes = context["children"]
        # should now include both P10 and P11
        self.assertEqual(len(nodes[1].children), 1)
        self.assertEqual(len(nodes[1].children[0].children), 1)

    def test_show_submenu_template_root_level_none_no_nephew_limit(self):
        root = self.get_page(1)
        context = self.get_context(path=root.get_absolute_url(), page=root)
        tpl = Template("{% load menu_tags %}{% show_sub_menu 100 None 100 %}")
        tpl.render(context)
        nodes = context["children"]
        # default nephew limit, P2 and P9 in the nodes list
        self.assertEqual(len(nodes), 2)


class FixturesMenuTests(MenusFixture, BaseMenuTest):
    """
    Tree from fixture:

        + P1
        | + P2
        |   + P3
        + P4
        | + P5
        + P6 (not in menu)
          + P7
          + P8
    """
    def get_page(self, num):
        return Page.objects.get(pagecontent_set__title='P%s' % num)

    def get_all_pages(self):
        return Page.objects.all()

    def test_menu_failfast_on_invalid_usage(self):
        context = self.get_context()
        context['child'] = self.get_page(1)
        # test standard show_menu
        with self.settings(DEBUG=True, TEMPLATE_DEBUG=True):
            tpl = Template("{% load menu_tags %}{% show_menu 0 0 0 0 'menu/menu.html' child %}")
            self.assertRaises(TemplateSyntaxError, tpl.render, context)

    def test_basic_cms_menu(self):
        menus = menu_pool.get_registered_menus()
        self.assertEqual(len(menus), 1)
        with force_language("en"):
            response = self.client.get(self.get_pages_root())  # path = '/'
        self.assertEqual(response.status_code, 200)
        request = self.get_request()

        renderer = menu_pool.get_renderer(request)

        # test the cms menu class
        menu = renderer.get_menu('CMSMenu')
        nodes = menu.get_nodes(request)
        pages = self.get_all_pages().order_by('node__path')
        self.assertEqual(len(nodes), len(pages))
        self.assertSequenceEqual(
            [node.get_absolute_url() for node in nodes],
            [page.get_absolute_url() for page in pages],
        )

    def test_show_page_in_menu_after_move_page(self):
        """
        Test checks if the menu cache is cleaned after move page.
        """
        page = create_page('page to move', 'nav_playground.html', 'en')

        request = self.get_request('/')
        renderer = menu_pool.get_renderer(request)
        renderer.draft_mode_active = True
        nodes_before = renderer.get_nodes()
        index_before = [i for i, s in enumerate(nodes_before) if s.title == page.get_title()]
        self.assertEqual(CacheKey.objects.count(), 1)

        with self.login_user_context(self.get_superuser()):
            # Moves the page to the second position in the tree
            data = {'id': page.pk, 'position': 1}
            endpoint = self.get_admin_url(Page, 'move_page', page.pk)
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(CacheKey.objects.count(), 0)

        request = self.get_request('/')
        renderer = menu_pool.get_renderer(request)
        renderer.draft_mode_active = True
        nodes_after = renderer.get_nodes()
        index_after = [i for i, s in enumerate(nodes_after) if s.title == page.get_title()]

        self.assertEqual(CacheKey.objects.count(), 1)
        self.assertNotEqual(
            index_before,
            index_after,
            'Index should not be the same after move page in navigation'
        )

    def test_cms_menu_public_with_multiple_languages(self):
        pages = self.get_all_pages().order_by('node__path')

        # Fallbacks on
        request = self.get_request(path='/de/', language='de')
        renderer = menu_pool.get_renderer(request)
        menu = renderer.get_menu('CMSMenu')
        nodes = menu.get_nodes(request)
        self.assertEqual(len(nodes), len(pages))

        with force_language('de'):
            # Current language is "de" but urls should still point
            # to "en" because of fallbacks.
            self.assertSequenceEqual(
                [node.get_absolute_url() for node in nodes],
                [page.get_absolute_url('en', fallback=False) for page in pages],
            )

        # Fallbacks off
        request = self.get_request(path='/de/', language='de')
        lang_settings = copy.deepcopy(get_cms_setting('LANGUAGES'))
        lang_settings[1][1]['hide_untranslated'] = True

        with self.settings(CMS_LANGUAGES=lang_settings):
            renderer = menu_pool.get_renderer(request)
            menu = renderer.get_menu('CMSMenu')
            nodes = menu.get_nodes(request)
            self.assertEqual(len(nodes), 0)

        for page in Page.objects.all():
            create_page_content(
                language='de',
                title=page.get_title('en'),
                page=page,
                slug='{}-de'.format(page.get_slug('en'))
            )

        # Fallbacks on
        # This time however, the "de" translations exist.
        request = self.get_request(path='/de/', language='de')
        renderer = menu_pool.get_renderer(request)
        menu = renderer.get_menu('CMSMenu')
        nodes = menu.get_nodes(request)
        self.assertEqual(len(nodes), len(pages))

        with force_language('de'):
            self.assertSequenceEqual(
                [node.get_absolute_url() for node in nodes],
                [page.get_absolute_url('de', fallback=False) for page in pages],
            )

        # Fallbacks off
        request = self.get_request(path='/de/', language='de')

        with self.settings(CMS_LANGUAGES=lang_settings):
            renderer = menu_pool.get_renderer(request)
            menu = renderer.get_menu('CMSMenu')
            nodes = menu.get_nodes(request)

            with force_language('de'):
                self.assertSequenceEqual(
                    [node.get_absolute_url() for node in nodes],
                    [page.get_absolute_url('de', fallback=False) for page in pages],
                )

    def test_show_menu(self):
        root = self.get_page(1)
        context = self.get_context(page=root)
        # test standard show_menu
        tpl = Template("{% load menu_tags %}{% show_menu %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 2)
        self.assertEqual(nodes[0].selected, True)
        self.assertEqual(nodes[0].sibling, False)
        self.assertEqual(nodes[0].descendant, False)
        self.assertEqual(nodes[0].children[0].descendant, True)
        self.assertEqual(nodes[0].children[0].children[0].descendant, True)
        self.assertEqual(nodes[0].get_absolute_url(), self.get_pages_root())
        self.assertEqual(nodes[1].get_absolute_url(), self.get_page(4).get_absolute_url())
        self.assertEqual(nodes[1].sibling, True)
        self.assertEqual(nodes[1].selected, False)

    def test_show_menu_num_queries(self):
        context = self.get_context()
        # test standard show_menu
        with self.assertNumQueries(6):
            """
            The queries should be:
                get all page nodes
                get all page permissions
                get all titles
                get all page urls
                get the menu cache key
                set the menu cache key
            """
            tpl = Template("{% load menu_tags %}{% show_menu %}")
            tpl.render(context)

    def test_show_menu_cache_key_leak(self):
        context = self.get_context()
        tpl = Template("{% load menu_tags %}{% show_menu %}")
        self.assertEqual(CacheKey.objects.count(), 0)
        tpl.render(context)
        self.assertEqual(CacheKey.objects.count(), 1)
        tpl.render(context)
        self.assertEqual(CacheKey.objects.count(), 1)

    def test_menu_cache_respects_database_keys(self):
        cms_page = self.get_page(1)
        context = self.get_context(path=cms_page.get_absolute_url(), page=cms_page)
        context['request'].session['cms_edit'] = False

        # Prime the cache
        with self.assertNumQueries(6):
            # The queries should be:
            #     get all page nodes
            #     get all page permissions
            #     get all titles
            #     get all page urls
            #     get the menu cache key
            #     set the menu cache key
            Template("{% load menu_tags %}{% show_menu %}").render(context)

        # One new CacheKey should have been created
        self.assertEqual(CacheKey.objects.count(), 1)

        # Because its cached, only one query is made to the db
        with self.assertNumQueries(1):
            # The queries should be:
            #     get the menu cache key
            Template("{% load menu_tags %}{% show_menu %}").render(context)

        # Delete the current cache key but don't touch the cache
        CacheKey.objects.all().delete()

        # The menu should be recalculated
        with self.assertNumQueries(6):
            # The queries should be:
            #     check if cache key exists
            #     get all page nodes
            #     get all page permissions
            #     get all title objects
            #     get all page url objects
            #     set the menu cache key
            Template("{% load menu_tags %}{% show_menu %}").render(context)

    def test_menu_keys_duplicate_clear(self):
        """
        Tests that the menu clears all keys, including duplicates.
        """
        CacheKey.objects.create(language="fr", site=1, key="a")
        CacheKey.objects.create(language="fr", site=1, key="a")

        self.assertEqual(CacheKey.objects.count(), 2)
        menu_pool.clear(site_id=1, language='fr')
        self.assertEqual(CacheKey.objects.count(), 0)

    def test_only_active_tree(self):
        context = self.get_context(page=self.get_page(1))
        # test standard show_menu
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 0 100 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes[1].children), 0)
        self.assertEqual(len(nodes[0].children), 1)
        self.assertEqual(len(nodes[0].children[0].children), 1)

        page_4 = self.get_page(4)
        context = self.get_context(path=page_4.get_absolute_url(), page=page_4)
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 0 100 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes[1].children), 1)
        self.assertEqual(len(nodes[0].children), 0)

    def test_only_one_active_level(self):
        context = self.get_context(page=self.get_page(1))
        # test standard show_menu
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 0 1 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes[1].children), 0)
        self.assertEqual(len(nodes[0].children), 1)
        self.assertEqual(len(nodes[0].children[0].children), 0)

    def test_only_level_zero(self):
        context = self.get_context()
        # test standard show_menu
        tpl = Template("{% load menu_tags %}{% show_menu 0 0 0 0 %}")
        tpl.render(context)
        nodes = context['children']
        for node in nodes:
            self.assertEqual(len(node.children), 0)

    def test_only_level_one(self):
        site = get_current_site()
        context = self.get_context()
        # test standard show_menu
        tpl = Template("{% load menu_tags %}{% show_menu 1 1 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        level_2_pages = Page.objects.filter(node__depth=2, node__site=site)
        self.assertEqual(len(nodes), level_2_pages.count())
        for node in nodes:
            self.assertEqual(len(node.children), 0)

    def test_only_level_one_active(self):
        context = self.get_context(page=self.get_page(1))
        # test standard show_menu
        tpl = Template("{% load menu_tags %}{% show_menu 1 1 0 100 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].descendant, True)
        self.assertEqual(len(nodes[0].children), 0)

    def test_level_zero_and_one(self):
        context = self.get_context()
        # test standard show_menu
        tpl = Template("{% load menu_tags %}{% show_menu 0 1 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 2)
        for node in nodes:
            self.assertEqual(len(node.children), 1)

    def test_show_submenu(self):
        context = self.get_context(page=self.get_page(1))
        # test standard show_menu
        tpl = Template("{% load menu_tags %}{% show_sub_menu %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(nodes[0].descendant, True)
        self.assertEqual(len(nodes), 1)
        self.assertEqual(len(nodes[0].children), 1)

        tpl = Template("{% load menu_tags %}{% show_sub_menu 1  %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 1)
        self.assertEqual(len(nodes[0].children), 0)

        page_3 = self.get_page(3)
        context = self.get_context(path=page_3.get_absolute_url(), page=page_3)
        tpl = Template("{% load menu_tags %}{% show_sub_menu 100 1 %}")
        tpl.render(context)
        nodes = context["children"]
        # P3 is the selected node
        self.assertFalse(nodes[0].selected)
        self.assertTrue(nodes[0].children[0].selected)
        # top level node should be P2
        self.assertEqual(nodes[0].get_absolute_url(), self.get_page(2).get_absolute_url())
        # should include P3 as well
        self.assertEqual(len(nodes[0].children), 1)

        page_2 = self.get_page(2)
        context = self.get_context(path=page_2.get_absolute_url(), page=page_2)
        tpl = Template("{% load menu_tags %}{% show_sub_menu 100 0 %}")
        tpl.render(context)
        nodes = context["children"]
        # P1 should be in the nav
        self.assertEqual(nodes[0].get_absolute_url(), self.get_page(1).get_absolute_url())
        # P2 is selected
        self.assertTrue(nodes[0].children[0].selected)

    def test_show_submenu_template_root_level_none(self):
        root = self.get_page(1)
        context = self.get_context(path=root.get_absolute_url(), page=root)
        tpl = Template("{% load menu_tags %}{% show_sub_menu 100 None 1 %}")
        tpl.render(context)
        nodes = context["children"]
        # First node is P2 (P1 children) thus not selected
        self.assertFalse(nodes[0].selected)
        # nephew limit of 1, so only P2 is the nodes list
        self.assertEqual(len(nodes), 1)
        # P3 is a child of P2, but not in nodes list
        self.assertTrue(nodes[0].children)

    def test_show_breadcrumb(self):
        page_3 = self.get_page(3)
        context = self.get_context(path=self.get_page(3).get_absolute_url(), page=page_3)
        tpl = Template("{% load menu_tags %}{% show_breadcrumb %}")
        tpl.render(context)
        nodes = context['ancestors']
        self.assertEqual(len(nodes), 3)
        tpl = Template("{% load menu_tags %}{% show_breadcrumb 1 %}")
        tpl.render(context)
        nodes = context['ancestors']
        self.assertEqual(len(nodes), 2)
        context = self.get_context()
        tpl = Template("{% load menu_tags %}{% show_breadcrumb %}")
        tpl.render(context)
        nodes = context['ancestors']
        self.assertEqual(len(nodes), 1)
        tpl = Template("{% load menu_tags %}{% show_breadcrumb 1 %}")
        tpl.render(context)
        nodes = context['ancestors']
        self.assertEqual(len(nodes), 0)

        page1 = self.get_page(1)
        page1.in_navigation = False
        page1.save()
        page2 = self.get_page(2)
        context = self.get_context(path=page2.get_absolute_url(), page=page2)
        tpl = Template("{% load menu_tags %}{% show_breadcrumb %}")
        tpl.render(context)
        nodes = context['ancestors']
        self.assertEqual(len(nodes), 2)
        self.assertEqual(nodes[0].get_absolute_url(), self.get_pages_root())
        self.assertEqual(isinstance(nodes[0], NavigationNode), True)
        self.assertEqual(nodes[1].get_absolute_url(), page2.get_absolute_url())

    def test_language_chooser(self):
        # test simple language chooser with default args
        lang_settings = copy.deepcopy(get_cms_setting('LANGUAGES'))
        lang_settings[1][0]['public'] = False
        with self.settings(CMS_LANGUAGES=lang_settings):
            context = self.get_context(path=self.get_page(3).get_absolute_url())
            tpl = Template("{% load menu_tags %}{% language_chooser %}")
            tpl.render(context)
            self.assertEqual(len(context['languages']), 3)
            # try a different template and some different args
            tpl = Template("{% load menu_tags %}{% language_chooser 'menu/test_language_chooser.html' %}")
            tpl.render(context)
            self.assertEqual(context['template'], 'menu/test_language_chooser.html')
            tpl = Template("{% load menu_tags %}{% language_chooser 'short' 'menu/test_language_chooser.html' %}")
            tpl.render(context)
            self.assertEqual(context['template'], 'menu/test_language_chooser.html')
            for lang in context['languages']:
                self.assertEqual(*lang)

    def test_language_chooser_all_for_staff(self):
        """
        Language chooser should show all configured languages
        on the current site if the user is staff.
        """
        superuser = self.get_superuser()
        lang_settings = copy.deepcopy(get_cms_setting('LANGUAGES'))
        # DE is marked as public False
        lang_settings[1][1]['public'] = False
        # FR is marked as public False
        lang_settings[1][2]['public'] = False

        with self.settings(CMS_LANGUAGES=lang_settings):
            with self.login_user_context(superuser):
                context = self.get_context(path=self.get_page(3).get_absolute_url())
                Template("{% load menu_tags %}{% language_chooser %}").render(context)
                self.assertEqual(len(context['languages']), 5)
                self.assertSequenceEqual(
                    sorted(lang[0] for lang in context['languages']),
                    ['de', 'en', 'es-mx', 'fr', 'pt-br']
                )

    def test_language_chooser_public_for_anon(self):
        """
        Language chooser should only show public configured languages
        on the current site if the user is anon.
        """
        # PT-BR is already set to public False
        lang_settings = copy.deepcopy(get_cms_setting('LANGUAGES'))
        # DE is marked as public False
        lang_settings[1][1]['public'] = False
        # FR is marked as public False
        lang_settings[1][2]['public'] = False

        with self.settings(CMS_LANGUAGES=lang_settings):
            context = self.get_context(path=self.get_page(3).get_absolute_url())
            Template("{% load menu_tags %}{% language_chooser %}").render(context)
            self.assertEqual(len(context['languages']), 2)
            self.assertSequenceEqual(
                sorted(lang[0] for lang in context['languages']),
                ['en', 'es-mx']
            )

    def test_page_language_url(self):
        path = self.get_page(3).get_absolute_url()
        context = self.get_context(path=path)
        tpl = Template("{%% load menu_tags %%}{%% page_language_url '%s' %%}" % 'en')
        url = tpl.render(context)
        self.assertEqual(url, "%s" % path)

    def test_show_menu_below_id(self):
        page2 = self.get_page(2)
        page2.reverse_id = "hello"
        page2.save()
        page2 = self.reload(page2)
        self.assertEqual(page2.reverse_id, "hello")
        page5 = self.get_page(5)
        context = self.get_context(path=page5.get_absolute_url())
        tpl = Template("{% load menu_tags %}{% show_menu_below_id 'hello' %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 1)
        page3_url = self.get_page(3).get_absolute_url()
        self.assertEqual(nodes[0].get_absolute_url(), page3_url)
        page2.in_navigation = False
        page2.save()
        context = self.get_context(path=page5.get_absolute_url())
        tpl = Template("{% load menu_tags %}{% show_menu_below_id 'hello' %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].get_absolute_url(), page3_url)

    def test_home_not_in_menu(self):
        page1 = self.get_page(1)
        page1.update_translations(in_navigation=False)
        page1.save()
        page4 = self.get_page(4)
        page4.update_translations(in_navigation=False)
        page4.save()
        context = self.get_context()
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].get_absolute_url(), self.get_page(2).get_absolute_url())
        self.assertEqual(nodes[0].children[0].get_absolute_url(), self.get_page(3).get_absolute_url())
        page4 = self.get_page(4)
        page4.update_translations(in_navigation=True)
        page4.save()
        menu_pool.clear(settings.SITE_ID)
        context = self.get_context()
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 2)

    def test_show_submenu_from_non_menu_page(self):
        """
        Here's the structure bit we're interested in:

        + P6 (not in menu)
          + P7
          + P8

        When we render P6, there should be a menu entry for P7 and P8 if the
        tag parameters are "1 XXX XXX XXX"
        """
        page6 = self.get_page(6)
        context = self.get_context(page6.get_absolute_url(), page=page6)
        tpl = Template("{% load menu_tags %}{% show_menu 1 100 0 1 %}")
        tpl.render(context)
        nodes = context['children']
        number_of_p6_children = page6.get_child_pages().filter(
            pagecontent_set__language='en', pagecontent_set__in_navigation=True).count()
        self.assertEqual(len(nodes), number_of_p6_children)

        page7 = self.get_page(7)
        context = self.get_context(page7.get_absolute_url(), page=page7)
        tpl = Template("{% load menu_tags %}{% show_menu 1 100 0 1 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), number_of_p6_children)

        tpl = Template("{% load menu_tags %}{% show_menu 2 100 0 1 %}")
        tpl.render(context)
        nodes = context['children']
        number_of_p7_children = page7.get_child_pages().filter(
            pagecontent_set__language='en', pagecontent_set__in_navigation=True).count()
        self.assertEqual(len(nodes), number_of_p7_children)

    def test_show_breadcrumb_invisible(self):
        parent = Page.objects.get(pagecontent_set__title='P3')
        invisible_page = create_page("invisible", "nav_playground.html", "en", parent=parent, in_navigation=False)
        context = self.get_context(
            path=invisible_page.get_absolute_url(),
            page=invisible_page,
        )
        tpl = Template("{% load menu_tags %}{% show_breadcrumb %}")
        tpl.render(context)
        nodes = context['ancestors']
        self.assertEqual(len(nodes), 3)
        tpl = Template("{% load menu_tags %}{% show_breadcrumb 0 'menu/breadcrumb.html' 1 %}")
        tpl.render(context)
        nodes = context['ancestors']
        self.assertEqual(len(nodes), 3)
        tpl = Template("{% load menu_tags %}{% show_breadcrumb 0 'menu/breadcrumb.html' 0 %}")
        tpl.render(context)
        nodes = context['ancestors']
        self.assertEqual(len(nodes), 4)


class MenuTests(BaseMenuTest):

    def test_build_nodes_inner_for_worst_case_menu(self):
        '''
            Tests the worst case scenario

            node5
             node4
              node3
               node2
                node1
        '''
        node1 = NavigationNode('Test1', '/test1/', 1, 2)
        node2 = NavigationNode('Test2', '/test2/', 2, 3)
        node3 = NavigationNode('Test3', '/test3/', 3, 4)
        node4 = NavigationNode('Test4', '/test4/', 4, 5)
        node5 = NavigationNode('Test5', '/test5/', 5, None)

        menu_class_name = 'Test'
        nodes = [node1, node2, node3, node4, node5, ]
        len_nodes = len(nodes)

        final_list = _build_nodes_inner_for_one_menu(nodes, menu_class_name)
        self.assertEqual(len(final_list), len_nodes)

        self.assertEqual(node1.parent, node2)
        self.assertEqual(node2.parent, node3)
        self.assertEqual(node3.parent, node4)
        self.assertEqual(node4.parent, node5)
        self.assertEqual(node5.parent, None)

        self.assertEqual(node1.children, [])
        self.assertEqual(node2.children, [node1])
        self.assertEqual(node3.children, [node2])
        self.assertEqual(node4.children, [node3])
        self.assertEqual(node5.children, [node4])

    def test_build_nodes_inner_for_circular_menu(self):
        '''
        TODO:
            To properly handle this test we need to have a circular dependency
            detection system.
            Go nuts implementing it :)
        '''
        pass

    def test_build_nodes_inner_for_broken_menu(self):
        '''
            Tests a broken menu tree (non-existing parent)

            node5
             node4
              node3

            <non-existant>
             node2
              node1
        '''
        node1 = NavigationNode('Test1', '/test1/', 1, 2)
        node2 = NavigationNode('Test2', '/test2/', 2, 12)
        node3 = NavigationNode('Test3', '/test3/', 3, 4)
        node4 = NavigationNode('Test4', '/test4/', 4, 5)
        node5 = NavigationNode('Test5', '/test5/', 5, None)

        menu_class_name = 'Test'
        nodes = [node1, node2, node3, node4, node5, ]

        final_list = _build_nodes_inner_for_one_menu(nodes, menu_class_name)
        self.assertEqual(len(final_list), 3)
        self.assertFalse(node1 in final_list)
        self.assertFalse(node2 in final_list)

        self.assertEqual(node1.parent, None)
        self.assertEqual(node2.parent, None)
        self.assertEqual(node3.parent, node4)
        self.assertEqual(node4.parent, node5)
        self.assertEqual(node5.parent, None)

        self.assertEqual(node1.children, [])
        self.assertEqual(node2.children, [])
        self.assertEqual(node3.children, [])
        self.assertEqual(node4.children, [node3])
        self.assertEqual(node5.children, [node4])

    def test_utils_mark_descendants(self):
        tree_nodes, flat_nodes = self._get_nodes()
        mark_descendants(tree_nodes)
        for node in flat_nodes:
            self.assertTrue(node.descendant, node)

    def test_utils_find_selected(self):
        tree_nodes, flat_nodes = self._get_nodes()
        node = flat_nodes[0]
        selected = find_selected(tree_nodes)
        self.assertEqual(selected, node)
        selected = find_selected([])
        self.assertEqual(selected, None)

    def test_utils_cut_levels(self):
        tree_nodes, flat_nodes = self._get_nodes()
        self.assertEqual(cut_levels(tree_nodes, 1), [flat_nodes[1]])

    def test_empty_menu(self):
        context = self.get_context()
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 0)

    def test_render_menu_with_invalid_language(self):
        """
        When rendering the menu, always fallback to a configured
        language on the current site.
        """
        # Refs - https://github.com/divio/django-cms/issues/6179
        site_2 = Site.objects.create(id=2, name='example-2.com', domain='example-2.com')
        de_defaults = {
            'site': site_2,
            'template': 'nav_playground.html',
            'language': 'de',
        }
        fr_defaults = {
            'site': site_2,
            'template': 'nav_playground.html',
            'language': 'fr',
        }
        create_page('DE-P1', in_navigation=True, **de_defaults)
        create_page('DE-P2', in_navigation=True, **de_defaults)
        create_page('DE-P3', in_navigation=True, **de_defaults)
        create_page('FR-P1', in_navigation=True, **fr_defaults)
        create_page('FR-P2', in_navigation=True, **fr_defaults)

        with self.settings(SITE_ID=2):
            request = self.get_request('/en/')
            context = Context()
            context['request'] = request
            tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
            tpl.render(context)
            nodes = context['children']
            self.assertEqual(len(nodes), 5)
            self.assertEqual(nodes[0].title, 'DE-P1')
            self.assertEqual(nodes[0].get_absolute_url(), '/de/de-p1/')
            self.assertEqual(nodes[1].title, 'DE-P2')
            self.assertEqual(nodes[1].get_absolute_url(), '/de/de-p2/')
            self.assertEqual(nodes[2].title, 'DE-P3')
            self.assertEqual(nodes[2].get_absolute_url(), '/de/de-p3/')
            self.assertEqual(nodes[3].title, 'FR-P1')
            self.assertEqual(nodes[3].get_absolute_url(), '/fr/fr-p1/')
            self.assertEqual(nodes[4].title, 'FR-P2')
            self.assertEqual(nodes[4].get_absolute_url(), '/fr/fr-p2/')

        menu_pool.clear(site_id=2)

        with self.settings(SITE_ID=2):
            request = self.get_request('/en/de-p2/')
            context = Context()
            context['request'] = request
            tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
            tpl.render(context)
            nodes = context['children']
            self.assertEqual(len(nodes), 5)
            self.assertEqual(nodes[0].title, 'DE-P1')
            self.assertEqual(nodes[0].get_absolute_url(), '/de/de-p1/')
            self.assertEqual(nodes[1].title, 'DE-P2')
            self.assertEqual(nodes[1].get_absolute_url(), '/de/de-p2/')
            self.assertEqual(nodes[2].title, 'DE-P3')
            self.assertEqual(nodes[2].get_absolute_url(), '/de/de-p3/')
            self.assertEqual(nodes[3].title, 'FR-P1')
            self.assertEqual(nodes[3].get_absolute_url(), '/fr/fr-p1/')
            self.assertEqual(nodes[4].title, 'FR-P2')
            self.assertEqual(nodes[4].get_absolute_url(), '/fr/fr-p2/')

    def test_render_menu_with_invalid_language_and_page(self):
        """
        This tests an edge-case where the user requests a
        language not configure for the current site
        while having pages on the current site with unconfigured
        translations.
        """
        # Refs - https://github.com/divio/django-cms/issues/6179
        site_2 = Site.objects.create(id=2, name='example-2.com', domain='example-2.com')
        de_defaults = {
            'site': site_2,
            'template': 'nav_playground.html',
            'language': 'de',
            'in_navigation': True,
        }
        create_page('DE-P1', **de_defaults)
        create_page('DE-P2', **de_defaults)
        create_page('DE-P3', **de_defaults)

        # The nl language is not configured for the current site
        # as a result, we have to create the pages manually.
        nl_page_1 = Page()
        nl_page_1.set_tree_node(site=site_2, target=None)
        nl_page_1.save()
        nl_page_1.pagecontent_set.create(
            language='nl',
            title='NL-P1',
            template='nav_playground.html',
            in_navigation=True,
        )

        nl_page_2 = Page()
        nl_page_2.set_tree_node(site=site_2, target=None)
        nl_page_2.save()
        nl_page_2.pagecontent_set.create(
            language='nl',
            title='NL-P2',
            template='nav_playground.html',
            in_navigation=True,
        )
        create_page_content('fr', 'FR-P2', nl_page_2, in_navigation=True)

        with self.settings(SITE_ID=2):
            request = self.get_request('/en/')
            context = Context()
            context['request'] = request
            tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
            tpl.render(context)
            nodes = context['children']
            self.assertEqual(len(nodes), 4)
            self.assertEqual(nodes[0].title, 'DE-P1')
            self.assertEqual(nodes[0].get_absolute_url(), '/de/de-p1/')
            self.assertEqual(nodes[1].title, 'DE-P2')
            self.assertEqual(nodes[1].get_absolute_url(), '/de/de-p2/')
            self.assertEqual(nodes[2].title, 'DE-P3')
            self.assertEqual(nodes[2].get_absolute_url(), '/de/de-p3/')
            self.assertEqual(nodes[3].title, 'FR-P2')
            self.assertEqual(nodes[3].get_absolute_url(), '/fr/fr-p2/')

        menu_pool.clear(site_id=2)

        with self.settings(SITE_ID=2):
            request = self.get_request('/en/de-p2/')
            context = Context()
            context['request'] = request
            tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
            tpl.render(context)
            nodes = context['children']
            self.assertEqual(len(nodes), 4)
            self.assertEqual(nodes[0].title, 'DE-P1')
            self.assertEqual(nodes[0].get_absolute_url(), '/de/de-p1/')
            self.assertEqual(nodes[1].title, 'DE-P2')
            self.assertEqual(nodes[1].get_absolute_url(), '/de/de-p2/')
            self.assertEqual(nodes[2].title, 'DE-P3')
            self.assertEqual(nodes[2].get_absolute_url(), '/de/de-p3/')
            self.assertEqual(nodes[3].title, 'FR-P2')
            self.assertEqual(nodes[3].get_absolute_url(), '/fr/fr-p2/')

    def test_render_menu_with_invalid_language_and_no_fallbacks(self):
        """
        The requested language is valid but there's no page
        with it and the user has disabled all fallbacks.
        The cms should render only nodes for the requested language.
        """
        defaults = {
            'template': 'nav_playground.html',
            'language': 'de',
        }
        create_page('DE-P1', in_navigation=True, **defaults)
        create_page('DE-P2', in_navigation=True, **defaults)
        create_page('DE-P3', in_navigation=True, **defaults)

        lang_settings = copy.deepcopy(get_cms_setting('LANGUAGES'))
        lang_settings[1][0]['fallbacks'] = []
        lang_settings[1][1]['fallbacks'] = []

        with self.settings(CMS_LANGUAGES=lang_settings):
            request = self.get_request('/en/')
            context = Context()
            context['request'] = request
            tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
            tpl.render(context)
            nodes = context['children']
            self.assertEqual(len(nodes), 0)

        menu_pool.clear(site_id=1)

        with self.settings(CMS_LANGUAGES=lang_settings):
            request = self.get_request('/en/de-p2/')
            context = Context()
            context['request'] = request
            tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
            tpl.render(context)
            nodes = context['children']
            self.assertEqual(len(nodes), 0)


@override_settings(CMS_PERMISSION=False)
class AdvancedSoftrootTests(SoftrootFixture, CMSTestCase):
    """
    Tree in fixture (as taken from issue 662):

        top
            root
                aaa
                    111
                        ccc
                            ddd
                    222
                bbb
                    333
                    444

    In the fixture, all pages are "in_navigation", "published" and
    NOT-"soft_root".

    What is a soft root?

        If a page is a soft root, it becomes the root page in the menu if
        we are currently on or under that page.

        If we are above that page, the children of this page are not shown.
    """

    def tearDown(self):
        Page.objects.all().delete()
        menu_pool.clear(all=True)

    def get_page(self, name):
        return Page.objects.get(urls__slug=name)

    def assertTreeQuality(self, a, b, *attrs):
        """
        Checks that the node-lists a and b are the same for attrs.

        This is recursive over the tree
        """
        msg = '%r != %r with %r, %r' % (len(a), len(b), a, b)
        self.assertEqual(len(a), len(b), msg)
        for n1, n2 in zip(a, b):
            for attr in attrs:
                a1 = getattr(n1, attr)
                a2 = getattr(n2, attr)
                msg = '%r != %r with %r, %r (%s)' % (a1, a2, n1, n2, attr)
                self.assertEqual(a1, a2, msg)
            self.assertTreeQuality(n1.children, n2.children)

    def test_top_not_in_nav(self):
        """
        top: not in navigation

        tag: show_menu 0 100 0 100

        context shared: current page is aaa
        context 1: root is NOT a softroot
        context 2: root IS a softroot

        expected result: the two node-trees should be equal
        """
        top = self.get_page('top')
        top.in_navigation = False
        top.save()
        aaa = self.get_page('aaa')
        # root is NOT a soft root
        context = self.get_context(aaa.get_absolute_url())
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 0 100 %}")
        tpl.render(context)
        hard_root = context['children']
        # root IS a soft root
        root = self.get_page('root')
        root.update_translations(soft_root=True)
        aaa = self.get_page('aaa')
        context = self.get_context(aaa.get_absolute_url())
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 0 100 %}")
        tpl.render(context)
        soft_root = context['children']
        # assert the two trees are equal in terms of 'level' and 'title'
        self.assertTreeQuality(hard_root, soft_root, 'level', 'title')

    def test_menu_tree_without_soft_root(self):
        """
        tag: show_menu 0 100 0 100

        expected result 1:
            0:top
               1:root
                  2:aaa
                     3:111
                        4:ccc
                           5:ddd
                     3:222
                  2:bbb
        """
        aaa = self.get_page('aaa')
        # root is NOT a soft root
        context = self.get_context(aaa.get_absolute_url(), page=aaa)
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 0 100 %}")
        tpl.render(context)
        hard_root = context['children']
        mock_tree = [
            AttributeObject(title='top', level=0, children=[
                AttributeObject(title='root', level=1, children=[
                    AttributeObject(title='aaa', level=2, children=[
                        AttributeObject(title='111', level=3, children=[
                            AttributeObject(title='ccc', level=4, children=[
                                AttributeObject(title='ddd', level=5, children=[])
                            ])
                        ]),
                        AttributeObject(title='222', level=3, children=[])
                    ]),
                    AttributeObject(title='bbb', level=2, children=[])
                ])
            ])
        ]
        self.assertTreeQuality(hard_root, mock_tree)

    def test_menu_tree_with_soft_root(self):
        """
        tag: show_menu 0 100 0 100

        expected result 2:
            0:root
               1:aaa
                  2:111
                     3:ccc
                        4:ddd
                  2:222
               1:bbb
        """
        # root IS a soft root
        root = self.get_page('root')
        root.update_translations(soft_root=True)
        aaa = self.get_page('aaa')
        context = self.get_context(aaa.get_absolute_url(), page=aaa)
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 0 100 %}")
        tpl.render(context)
        soft_root = context['children']
        mock_tree = [
            AttributeObject(title='root', level=0, children=[
                AttributeObject(title='aaa', level=1, children=[
                    AttributeObject(title='111', level=2, children=[
                        AttributeObject(title='ccc', level=3, children=[
                            AttributeObject(title='ddd', level=4, children=[])
                        ])
                    ]),
                    AttributeObject(title='222', level=2, children=[])
                ]),
                AttributeObject(title='bbb', level=1, children=[])
            ])
        ]
        self.assertTreeQuality(soft_root, mock_tree, 'title', 'level')


class ShowSubMenuCheck(SubMenusFixture, BaseMenuTest):
    """
    Tree from fixture:

        + P1
        | + P2
        |   + P3
        + P4
        | + P5
        + P6
          + P7 (not in menu)
          + P8
    """
    def test_show_submenu(self):
        page = self.get_page(6)
        subpage = self.get_page(8)
        context = self.get_context(page.get_absolute_url(), page=page)
        # test standard show_menu
        tpl = Template("{% load menu_tags %}{% show_sub_menu %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].id, subpage.pk)

    def test_show_submenu_num_queries(self):
        page = self.get_page(6)
        subpage = self.get_page(8)
        context = self.get_context(page.get_absolute_url(), page=page)

        # test standard show_menu
        with self.assertNumQueries(6):
            """
            The queries should be:
                get all page nodes
                get all page permissions
                get all titles
                get all page urls
                get the menu cache key
                set the menu cache key
            """
            tpl = Template("{% load menu_tags %}{% show_sub_menu %}")
            tpl.render(context)
            nodes = context['children']
            self.assertEqual(len(nodes), 1)
            self.assertEqual(nodes[0].id, subpage.pk)


class ShowMenuBelowIdTests(BaseMenuTest):
    """
    Test for issue 521

    Build the following tree:

        A
        |-B
          |-C
          \\-D (not in nav)
    """
    def test_not_in_navigation(self):
        a = create_page('A', 'nav_playground.html', 'en', in_navigation=True, reverse_id='a')
        b = create_page('B', 'nav_playground.html', 'en', parent=a, in_navigation=True)
        c = create_page('C', 'nav_playground.html', 'en', parent=b, in_navigation=True)
        create_page('D', 'nav_playground.html', 'en', parent=self.reload(b), in_navigation=False)
        context = self.get_context(a.get_absolute_url())
        tpl = Template("{% load menu_tags %}{% show_menu_below_id 'a' 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 1, nodes)
        node = nodes[0]
        self.assertEqual(node.id, b.pk)
        children = node.children
        self.assertEqual(len(children), 1, repr(children))
        child = children[0]
        self.assertEqual(child.id, c.pk)

    def test_menu_beyond_soft_root(self):
        """
        Test for issue 4107

        Build the following tree:

            A
            |-B (soft_root)
              |-C
        """
        stdkwargs = {
            'template': 'nav_playground.html',
            'language': 'en',
            'in_navigation': True,
        }
        a = create_page('A', reverse_id='a', **stdkwargs)
        b = create_page('B', parent=a, soft_root=True, **stdkwargs)
        c = create_page('C', parent=b, **stdkwargs)

        context = self.get_context(a.get_absolute_url(), page=a)
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check whole menu
        self.assertEqual(len(nodes), 1)
        a_node = nodes[0]
        self.assertEqual(a_node.id, a.pk)  # On A, show from A
        self.assertEqual(len(a_node.children), 1)
        b_node = a_node.children[0]
        self.assertEqual(b_node.id, b.pk)
        self.assertEqual(len(b_node.children), 1)
        c_node = b_node.children[0]
        self.assertEqual(c_node.id, c.pk)
        self.assertEqual(len(c_node.children), 0)

        context = self.get_context(b.get_absolute_url(), page=b)
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check whole menu
        self.assertEqual(len(nodes), 1)
        b_node = nodes[0]
        self.assertEqual(b_node.id, b.pk)  # On B, show from B
        self.assertEqual(len(b_node.children), 1)
        c_node = b_node.children[0]
        self.assertEqual(c_node.id, c.pk)
        self.assertEqual(len(c_node.children), 0)

        context = self.get_context(c.get_absolute_url(), page=c)
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check whole menu
        self.assertEqual(len(nodes), 1)
        b_node = nodes[0]
        self.assertEqual(b_node.id, b.pk)  # On C, show from B
        self.assertEqual(len(b_node.children), 1)
        c_node = b_node.children[0]
        self.assertEqual(c_node.id, c.pk)
        self.assertEqual(len(c_node.children), 0)

        context = self.get_context(a.get_absolute_url(), page=a)
        tpl = Template("{% load menu_tags %}{% show_menu_below_id 'a' 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check whole menu
        self.assertEqual(len(nodes), 1)
        b_node = nodes[0]
        self.assertEqual(b_node.id, b.pk)  # On A, show from B (since below A)
        self.assertEqual(len(b_node.children), 1)
        c_node = b_node.children[0]
        self.assertEqual(c_node.id, c.pk)
        self.assertEqual(len(c_node.children), 0)

        context = self.get_context(b.get_absolute_url(), page=b)
        tpl = Template("{% load menu_tags %}{% show_menu_below_id 'a' 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check whole menu
        self.assertEqual(len(nodes), 1)
        b_node = nodes[0]
        self.assertEqual(b_node.id, b.pk)  # On B, show from B (since below A)
        self.assertEqual(len(b_node.children), 1)
        c_node = b_node.children[0]
        self.assertEqual(c_node.id, c.pk)
        self.assertEqual(len(c_node.children), 0)

        context = self.get_context(c.get_absolute_url(), page=c)
        tpl = Template("{% load menu_tags %}{% show_menu_below_id 'a' 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check whole menu
        self.assertEqual(len(nodes), 1)
        b_node = nodes[0]
        self.assertEqual(b_node.id, b.pk)  # On C, show from B (since below A)
        self.assertEqual(len(b_node.children), 1)
        c_node = b_node.children[0]
        self.assertEqual(c_node.id, c.pk)
        self.assertEqual(len(c_node.children), 0)

    def test_not_in_navigation_num_queries(self):
        """
        Test for issue 521

        Build the following tree:

            A
            |-B
              |-C
              \\-D (not in nav)
        """
        a = create_page('A', 'nav_playground.html', 'en',
                        in_navigation=True, reverse_id='a')
        b = create_page('B', 'nav_playground.html', 'en', parent=a,
                        in_navigation=True)
        c = create_page('C', 'nav_playground.html', 'en', parent=b,
                        in_navigation=True)
        create_page('D', 'nav_playground.html', 'en', parent=self.reload(b),
                    in_navigation=False)

        with LanguageOverride('en'):
            context = self.get_context(a.get_absolute_url())
            with self.assertNumQueries(6):
                """
                The queries should be:
                    get all page nodes
                    get all page permissions
                    get all titles
                    get all page urls
                    get the menu cache key
                    set the menu cache key
                """
                # Actually seems to run:
                tpl = Template("{% load menu_tags %}{% show_menu_below_id 'a' 0 100 100 100 %}")
                tpl.render(context)
            nodes = context['children']
            self.assertEqual(len(nodes), 1, nodes)
            node = nodes[0]
            self.assertEqual(node.id, b.pk)
            children = node.children
            self.assertEqual(len(children), 1, repr(children))
            child = children[0]
            self.assertEqual(child.id, c.pk)

    def test_menu_in_soft_root(self):
        """
        Test for issue 3504

        Build the following tree:

            A
            |-B
            C (soft_root)
        """
        a = create_page('A', 'nav_playground.html', 'en', in_navigation=True, reverse_id='a')
        b = create_page('B', 'nav_playground.html', 'en', parent=a, in_navigation=True)
        c = create_page('C', 'nav_playground.html', 'en', in_navigation=True, soft_root=True)
        context = self.get_context(a.get_absolute_url())
        tpl = Template("{% load menu_tags %}{% show_menu_below_id 'a' %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 1)
        node = nodes[0]
        self.assertEqual(node.id, b.pk)
        context = self.get_context(c.get_absolute_url())
        tpl = Template("{% load menu_tags %}{% show_menu_below_id 'a' %}")
        tpl.render(context)
        nodes = context['children']
        self.assertEqual(len(nodes), 1)
        node = nodes[0]
        self.assertEqual(node.id, b.pk)


@override_settings(
    CMS_PERMISSION=True,
    CMS_PUBLIC_FOR='staff',
)
class ViewPermissionMenuTests(CMSTestCase):

    def setUp(self):
        self.page = create_page('page', 'nav_playground.html', 'en')
        self.pages = [self.page]
        self.user = self.get_standard_user()
        self.site = get_current_site()

    def get_request(self, user=None):
        attrs = {
            'user': user or AnonymousUser(),
            'REQUEST': {},
            'POST': {},
            'GET': {},
            'session': {},
        }
        return type('Request', (object,), attrs)

    def test_public_for_all_staff(self):
        request = self.get_request(self.user)
        request.user.is_staff = True
        with self.assertNumQueries(4):
            """
            The queries are:
            User permissions
            Content type
            GlobalPagePermission query
            PagePermission count query
            """
            pages = get_visible_nodes(request, self.pages, self.site)
            self.assertEqual(pages, self.pages)

    @override_settings(CMS_PUBLIC_FOR='all')
    def test_public_for_all(self):
        request = self.get_request(self.user)

        with self.assertNumQueries(4):
            """
            The queries are:
            User permissions
            Content type
            GlobalPagePermission query
            PagePermission query for affected pages
            """
            pages = get_visible_nodes(request, self.pages, self.site)
            self.assertEqual(pages, self.pages)

    @override_settings(CMS_PUBLIC_FOR='all')
    def test_unauthed(self):
        request = self.get_request()
        with self.assertNumQueries(1):
            """
            The query is:
            PagePermission query for affected pages
            """
            pages = get_visible_nodes(request, self.pages, self.site)
            self.assertEqual(pages, self.pages)

    def test_authed_basic_perm(self):
        self.user.user_permissions.add(Permission.objects.get(codename='view_page'))
        request = self.get_request(self.user)

        with self.assertNumQueries(2):
            """
            The queries are:
            User permissions
            Content type
            """
            pages = get_visible_nodes(request, self.pages, self.site)
            self.assertEqual(pages, self.pages)

    def test_authed_no_access(self):
        request = self.get_request(self.user)

        with self.assertNumQueries(4):
            """
            The queries are:
            View Permission Calculation Query
            GlobalpagePermission query for user
            User permissions
            Content type
            """
            pages = get_visible_nodes(request, self.pages, self.site)
            self.assertEqual(pages, [])

    def test_unauthed_no_access(self):
        request = self.get_request()

        with self.assertNumQueries(0):
            nodes = get_visible_nodes(request, self.pages, self.site)
            self.assertEqual(nodes, [])

    def test_page_permissions(self):
        request = self.get_request(self.user)
        PagePermission.objects.create(can_view=True, user=self.user, page=self.page)

        with self.assertNumQueries(4):
            """
            The queries are:
            PagePermission query for affected pages
            User permissions
            Content type
            GlobalpagePermission query for user
            """
            pages = get_visible_nodes(request, self.pages, self.site)
            self.assertEqual(pages, self.pages)

    def test_page_permissions_view_groups(self):
        group = Group.objects.create(name='testgroup')
        self.user.groups.add(group)
        request = self.get_request(self.user)
        PagePermission.objects.create(can_view=True, group=group, page=self.page)

        with self.assertNumQueries(5):
            """
            The queries are:
            PagePermission query for affected pages
            User permissions
            Content type
            GlobalpagePermission query for user
            Group query via PagePermission
            """
            pages = get_visible_nodes(request, self.pages, self.site)
            self.assertEqual(pages, self.pages)

    def test_global_permission(self):
        GlobalPagePermission.objects.create(can_view=True, user=self.user)
        request = self.get_request(self.user)
        group = Group.objects.create(name='testgroup')
        PagePermission.objects.create(can_view=True, group=group, page=self.page)

        with self.assertNumQueries(3):
            """
            The queries are:
            User permissions
            Content type
            GlobalpagePermission query for user
            """
            pages = get_visible_nodes(request, self.pages, self.site)
            self.assertEqual(pages, self.pages)


@override_settings(
    CMS_PERMISSION=True,
    CMS_PUBLIC_FOR='all',
)
class PublicViewPermissionMenuTests(CMSTestCase):

    def setUp(self):
        """
        Create this published hierarchy:
        A
        B1     B2
        C1 C2  C3 C4
        """
        template = 'nav_playground.html'
        kw = dict(in_navigation=True)
        a = create_page('a', template, 'en', **kw)
        b1 = create_page('b1', template, 'en', parent=a, **kw)
        b2 = create_page('b2', template, 'en', parent=a, **kw)
        c1 = create_page('c1', template, 'en', parent=b1, **kw)
        c2 = create_page('c2', template, 'en', parent=b1, **kw)
        c3 = create_page('c3', template, 'en', parent=b2, **kw)
        c4 = create_page('c4', template, 'en', parent=b2, **kw)
        self.pages = [a, b1, c1, c2, b2, c3, c4]  # tree order
        self.site = get_current_site()

        self.user = self._create_user("standard", is_staff=False, is_superuser=False)
        self.other = self._create_user("other", is_staff=False, is_superuser=False)
        PagePermission.objects.create(page=b1, user=self.user, can_view=True,
                                      grant_on=ACCESS_PAGE_AND_DESCENDANTS)
        PagePermission.objects.create(page=b2, user=self.other, can_view=True,
                                      grant_on=ACCESS_PAGE_AND_DESCENDANTS)
        attrs = {
            'user': self.user,
            'REQUEST': {},
            'POST': {},
            'GET': {},
            'session': {},
        }
        self.request = type('Request', (object,), attrs)

    def test_list_access(self):
        pages = get_visible_nodes(self.request, self.pages, self.site)
        pages = (
            Page
            .objects
            .filter(pk__in=(page.pk for page in pages))
            .values_list('pagecontent_set__title', flat=True)
        )
        self.assertSequenceEqual(sorted(pages), ['a', 'b1', 'c1', 'c2'])

    def test_qs_access(self):
        pages = get_visible_nodes(self.request, Page.objects.all(), self.site)
        pages = (
            Page
            .objects
            .filter(pk__in=(page.pk for page in pages))
            .values_list('pagecontent_set__title', flat=True)
        )
        self.assertSequenceEqual(sorted(pages), ['a', 'b1', 'c1', 'c2'])


@override_settings(CMS_PERMISSION=False)
class SoftrootTests(CMSTestCase):
    """
    Ask evildmp/superdmp if you don't understand softroots!

    Softroot description from the docs:

        A soft root is a page that acts as the root for a menu navigation tree.

        Typically, this will be a page that is the root of a significant new
        section on your site.

        When the soft root feature is enabled, the navigation menu for any page
        will start at the nearest soft root, rather than at the real root of
        the site’s page hierarchy.

        This feature is useful when your site has deep page hierarchies (and
        therefore multiple levels in its navigation trees). In such a case, you
        usually don’t want to present site visitors with deep menus of nested
        items.

        For example, you’re on the page “Introduction to Bleeding”, so the menu
        might look like this:

            School of Medicine
                Medical Education
                Departments
                    Department of Lorem Ipsum
                    Department of Donec Imperdiet
                    Department of Cras Eros
                    Department of Mediaeval Surgery
                        Theory
                        Cures
                        Bleeding
                            Introduction to Bleeding <this is the current page>
                            Bleeding - the scientific evidence
                            Cleaning up the mess
                            Cupping
                            Leaches
                            Maggots
                        Techniques
                        Instruments
                    Department of Curabitur a Purus
                    Department of Sed Accumsan
                    Department of Etiam
                Research
                Administration
                Contact us
                Impressum

        which is frankly overwhelming.

        By making “Department of Mediaeval Surgery” a soft root, the menu
        becomes much more manageable:

            Department of Mediaeval Surgery
                Theory
                Cures
                    Bleeding
                        Introduction to Bleeding <current page>
                        Bleeding - the scientific evidence
                        Cleaning up the mess
                    Cupping
                    Leaches
                    Maggots
                Techniques
                Instruments
    """

    def test_basic_home(self):
        """
        Given the tree:

        |- Home
        | |- Projects (SOFTROOT)
        | | |- django CMS
        | | |- django Shop
        | |- People

        Expected menu when on "Home" (0 100 100 100):

        |- Home
        | |- Projects (SOFTROOT)
        | | |- django CMS
        | | |- django Shop
        | |- People
        """
        stdkwargs = {
            'template': 'nav_playground.html',
            'language': 'en',
            'in_navigation': True,
        }
        home = create_page("Home", **stdkwargs)
        projects = create_page("Projects", parent=home, soft_root=True, **stdkwargs)
        djangocms = create_page("django CMS", parent=projects, **stdkwargs)
        djangoshop = create_page("django Shop", parent=projects, **stdkwargs)
        people = create_page("People", parent=home, **stdkwargs)
        # On Home
        context = self.get_context(home.get_absolute_url())
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check everything
        self.assertEqual(len(nodes), 1)
        homenode = nodes[0]
        self.assertEqual(homenode.id, home.pk)
        self.assertEqual(len(homenode.children), 2)
        projectsnode, peoplenode = homenode.children
        self.assertEqual(projectsnode.id, projects.pk)
        self.assertEqual(peoplenode.id, people.pk)
        self.assertEqual(len(projectsnode.children), 2)
        cmsnode, shopnode = projectsnode.children
        self.assertEqual(cmsnode.id, djangocms.pk)
        self.assertEqual(shopnode.id, djangoshop.pk)
        self.assertEqual(len(cmsnode.children), 0)
        self.assertEqual(len(shopnode.children), 0)
        self.assertEqual(len(peoplenode.children), 0)

    def test_basic_projects(self):
        """
        Given the tree:

        |- Home
        | |- Projects (SOFTROOT)
        | | |- django CMS
        | | |- django Shop
        | |- People

        Expected menu when on "Projects" (0 100 100 100):

        |- Projects (SOFTROOT)
        | |- django CMS
        | |- django Shop
        """
        stdkwargs = {
            'template': 'nav_playground.html',
            'language': 'en',
            'in_navigation': True,
        }
        home = create_page("Home", **stdkwargs)
        projects = create_page("Projects", parent=home, soft_root=True, **stdkwargs)
        djangocms = create_page("django CMS", parent=projects, **stdkwargs)
        djangoshop = create_page("django Shop", parent=projects, **stdkwargs)
        create_page("People", parent=home, **stdkwargs)
        # On Projects
        context = self.get_context(projects.get_absolute_url(), page=projects)
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check everything
        self.assertEqual(len(nodes), 1)
        projectsnode = nodes[0]
        self.assertEqual(projectsnode.id, projects.pk)
        self.assertEqual(len(projectsnode.children), 2)
        cmsnode, shopnode = projectsnode.children
        self.assertEqual(cmsnode.id, djangocms.pk)
        self.assertEqual(shopnode.id, djangoshop.pk)
        self.assertEqual(len(cmsnode.children), 0)
        self.assertEqual(len(shopnode.children), 0)

    def test_basic_djangocms(self):
        """
        Given the tree:

        |- Home
        | |- Projects (SOFTROOT)
        | | |- django CMS
        | | |- django Shop
        | |- People

        Expected menu when on "django CMS" (0 100 100 100):

        |- Projects (SOFTROOT)
        | |- django CMS
        | |- django Shop
        """
        stdkwargs = {
            'template': 'nav_playground.html',
            'language': 'en',
            'in_navigation': True,
        }
        home = create_page("Home", **stdkwargs)
        projects = create_page("Projects", parent=home, soft_root=True, **stdkwargs)
        djangocms = create_page("django CMS", parent=projects, **stdkwargs)
        djangoshop = create_page("django Shop", parent=projects, **stdkwargs)
        create_page("People", parent=home, **stdkwargs)
        # On django CMS
        context = self.get_context(djangocms.get_absolute_url(), page=djangocms)
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check everything
        self.assertEqual(len(nodes), 1)
        projectsnode = nodes[0]
        self.assertEqual(projectsnode.id, projects.pk)
        self.assertEqual(len(projectsnode.children), 2)
        cmsnode, shopnode = projectsnode.children
        self.assertEqual(cmsnode.id, djangocms.pk)
        self.assertEqual(shopnode.id, djangoshop.pk)
        self.assertEqual(len(cmsnode.children), 0)
        self.assertEqual(len(shopnode.children), 0)

    def test_basic_people(self):
        """
        Given the tree:

        |- Home
        | |- Projects (SOFTROOT)
        | | |- django CMS
        | | |- django Shop
        | |- People

        Expected menu when on "People" (0 100 100 100):

        |- Home
        | |- Projects (SOFTROOT)
        | | |- django CMS
        | | |- django Shop
        | |- People
        """
        stdkwargs = {
            'template': 'nav_playground.html',
            'language': 'en',
            'in_navigation': True,
        }
        home = create_page("Home", **stdkwargs)
        projects = create_page("Projects", parent=home, soft_root=True, **stdkwargs)
        djangocms = create_page("django CMS", parent=projects, **stdkwargs)
        djangoshop = create_page("django Shop", parent=projects, **stdkwargs)
        people = create_page("People", parent=home, **stdkwargs)
        # On People
        context = self.get_context(home.get_absolute_url())
        tpl = Template("{% load menu_tags %}{% show_menu 0 100 100 100 %}")
        tpl.render(context)
        nodes = context['children']
        # check everything
        self.assertEqual(len(nodes), 1)
        homenode = nodes[0]
        self.assertEqual(homenode.id, home.pk)
        self.assertEqual(len(homenode.children), 2)
        projectsnode, peoplenode = homenode.children
        self.assertEqual(projectsnode.id, projects.pk)
        self.assertEqual(peoplenode.id, people.pk)
        self.assertEqual(len(projectsnode.children), 2)
        cmsnode, shopnode = projectsnode.children
        self.assertEqual(cmsnode.id, djangocms.pk)
        self.assertEqual(shopnode.id, djangoshop.pk)
        self.assertEqual(len(cmsnode.children), 0)
        self.assertEqual(len(shopnode.children), 0)
        self.assertEqual(len(peoplenode.children), 0)



from django.template.loader import get_template
from django.utils.functional import cached_property


class TemplatesCache():

    def __init__(self):
        self._cached_templates = {}

    def get_cached_template(self, template):
        # we check if template quacks like a Template, as generic Template and engine-specific Template
        # does not share a common ancestor
        if hasattr(template, 'render'):
            return template

        if template not in self._cached_templates:
            # this always return an engine-specific template object
            self._cached_templates[template] = get_template(template)
        return self._cached_templates[template]

    @cached_property
    def drag_item_template(self):
        return get_template('cms/toolbar/dragitem.html')

    @cached_property
    def placeholder_plugin_menu_template(self):
        return get_template('cms/toolbar/dragitem_menu.html')

    @cached_property
    def dragbar_template(self):
        return get_template('cms/toolbar/dragbar.html')



from collections import defaultdict

from django.contrib.sitemaps import Sitemap
from django.db.models import Q

from cms.models import PageContent, PageUrl
from cms.utils import get_current_site
from cms.utils.i18n import get_public_languages


def from_iterable(iterables):
    """
    Backport of itertools.chain.from_iterable
    """
    for it in iterables:
        for element in it:
            yield element


class CMSSitemap(Sitemap):
    changefreq = "monthly"
    priority = 0.5

    def items(self):
        #
        # It is counter-productive to provide entries for:
        #   > Pages which redirect:
        #       - If the page redirects to another page on this site, the
        #         destination page will already be in the sitemap, and
        #       - If the page redirects externally, then it shouldn't be
        #         part of our sitemap anyway.
        #   > Pages which cannot be accessed by anonymous users (like
        #     search engines are).
        #
        # It is noted here: http://www.sitemaps.org/protocol.html that
        # "locations" that differ from the place where the sitemap is found,
        # are considered invalid. E.g., if your sitemap is located here:
        #
        #     http://example.com/sub/sitemap.xml
        #
        # valid locations *must* be rooted at http://example.com/sub/...
        #
        # This rules any redirected locations out.
        #
        # If, for some reason, you require redirecting pages (PageContent) to be
        # included, simply create a new class inheriting from this one, and
        # supply a new items() method which doesn't filter out the redirects.
        site = get_current_site()
        languages = get_public_languages(site_id=site.pk)
        all_urls = (
            PageUrl
            .objects
            .get_for_site(site)
            .select_related('page')
            .filter(language__in=languages, path__isnull=False, page__login_required=False)
            .order_by('page__node__path')
        )
        excluded_titles_by_page = defaultdict(set)
        excluded_translations = (
            PageContent
            .objects
            .filter(language__in=languages, page__node__site=site)
            .exclude(Q(redirect='') | Q(redirect__isnull=True))
            .values_list('page', 'language')
        )

        for page_id, language in excluded_translations:
            excluded_titles_by_page[page_id].add(language)

        valid_urls = []

        for page_url in all_urls:
            excluded = excluded_titles_by_page.get(page_url.page_id, [])

            if page_url.language in excluded:
                continue
            valid_urls.append(page_url)
        return valid_urls

    def lastmod(self, page_url):
        return page_url.page.changed_date

    def location(self, page_url):
        return page_url.get_absolute_url(page_url.language)



from django.http import HttpResponse

from cms.api import create_page, create_page_content
from cms.test_utils.testcases import CMSTestCase
from cms.test_utils.util.mock import AttributeObject
from cms.utils.i18n import get_language_list
from menus.templatetags.menu_tags import PageLanguageUrl
from menus.utils import (
    DefaultLanguageChanger,
    find_selected,
    language_changer_decorator,
)


class DumbPageLanguageUrl(PageLanguageUrl):
    def __init__(self):
        pass


class MenuUtilsTests(CMSTestCase):
    def get_simple_view(self):
        def myview(request):
            return HttpResponse('')
        return myview

    def test_reverse_in_changer(self):
        response = self.client.get('/en/sample/login/')
        self.assertContains(response, '<h1>/fr/sample/login/</h1>')

        response = self.client.get('/en/sample/login_other/')
        self.assertContains(response, '<h1>/fr/sample/login_other/</h1>')

        response = self.client.get('/en/sample/login3/')
        self.assertContains(response, '<h1>/fr/sample/login3/</h1>')

    def test_default_language_changer(self):
        """
        The DefaultLanguageChanger should not try to resolve the url
        for languages not configured.
        """
        cms_page = create_page('en-page', 'nav_playground.html', 'en')

        for language in get_language_list(site_id=1):
            if language not in ('en', 'pt-br', 'es-mx'):
                create_page_content(language, '%s-page' % language, cms_page)

        request = self.get_request(
            path=cms_page.get_absolute_url(),
            language='en',
            page=cms_page,
        )
        urls_expected = [
            '/en/en-page/',
            '/de/de-page/',
            '/fr/fr-page/',
            '/en/en-page/',  # the pt-br url is en because that's a fallback
            '/en/en-page/',  # the es-mx url is en because that's a fallback
        ]
        urls_found = [DefaultLanguageChanger(request)(code)
                      for code in get_language_list(site_id=1)]
        self.assertSequenceEqual(urls_expected, urls_found)

    def test_language_changer_decorator(self):
        def lang_changer(lang):
            return "/%s/dummy/" % lang
        decorated_view = language_changer_decorator(lang_changer)(self.get_simple_view())
        request = self.get_request('/some/path/', 'en')
        response = decorated_view(request)
        self.assertEqual(response.content, b'')
        fake_context = {'request': request}
        tag = DumbPageLanguageUrl()
        output = tag.get_context(fake_context, 'en')
        url = output['content']
        self.assertEqual(url, '/en/dummy/')
        output = tag.get_context(fake_context, 'ja')
        url = output['content']
        self.assertEqual(url, '/ja/dummy/')

    def test_find_selected(self):
        subchild = AttributeObject()
        firstchild = AttributeObject(ancestor=True, children=[subchild])
        selectedchild = AttributeObject(selected=True)
        secondchild = AttributeObject(ancestor=True, children=[selectedchild])
        root = AttributeObject(ancestor=True, children=[firstchild, secondchild])
        nodes = [root]
        selected = find_selected(nodes)
        self.assertEqual(selected, selectedchild)



from django.db import models

from cms.models import CMSPlugin

CLASS_CHOICES = ['container', 'content', 'teaser']
CLASS_CHOICES = tuple((entry, entry) for entry in CLASS_CHOICES)

TAG_CHOICES = [
    'div', 'article', 'section', 'header', 'footer', 'aside', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6'
]
TAG_CHOICES = tuple((entry, entry) for entry in TAG_CHOICES)


class Style(CMSPlugin):
    """
    Renders a given ``TAG_CHOICES`` element with additional attributes
    """
    label = models.CharField(
        verbose_name='Label',
        blank=True,
        max_length=255,
        help_text='Overrides the display name in the structure mode.',
    )
    tag_type = models.CharField(
        verbose_name='Tag type',
        choices=TAG_CHOICES,
        default=TAG_CHOICES[0][0],
        max_length=255,
    )
    class_name = models.CharField(
        verbose_name='Class name',
        choices=CLASS_CHOICES,
        default=CLASS_CHOICES[0][0],
        blank=True,
        max_length=255,
    )
    additional_classes = models.CharField(
        verbose_name='Additional classes',
        blank=True,
        max_length=255,
    )

    def __str__(self):
        return self.label or self.tag_type or str(self.pk)

    def get_short_description(self):
        # display format:
        # Style label <tag> .list.of.classes #id
        display = []
        classes = []

        if self.label:
            display.append(self.label)
        if self.tag_type:
            display.append('<{0}>'.format(self.tag_type))
        if self.class_name:
            classes.append(self.class_name)
        if self.additional_classes:
            classes.extend(item.strip() for item in self.additional_classes.split(',') if item.strip())
        display.append('.{0}'.format('.'.join(classes)))
        return ' '.join(display)

    def get_additional_classes(self):
        return ' '.join(item.strip() for item in self.additional_classes.split(',') if item.strip())



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin as OriginalUserAdmin
from django.contrib.auth.models import User as OriginalUser

if getattr(OriginalUser._meta, 'swapped', False):
    class UserAdmin(OriginalUserAdmin):
        list_display = ('username', 'email', 'get_full_name', 'is_staff')
        search_fields = ('username', 'email',)
        fieldsets = (
            (None, {'fields': ('username', 'password')}),
            ('Personal info', {'fields': ('email',)}),
            ('Permissions', {'fields': ('is_active', 'is_staff', 'is_superuser',
                                        'groups', 'user_permissions')}),
            ('Important dates', {'fields': ('last_login',)}),
        )

    admin.site.register(get_user_model(), UserAdmin)



from django.urls import include, re_path
from django.utils.translation import gettext_lazy as _

from . import views

"""
Also used in cms.tests.ApphooksTestCase
"""
urlpatterns = [
    re_path(r'^$', views.sample_view, {'message': 'sample root page', }, name='sample-root'),
    re_path(r'^exempt/$', views.exempt_view, {'message': 'sample root page', }, name='sample-exempt'),
    re_path(r'^settings/$', views.sample_view, kwargs={'message': 'sample settings page'}, name='sample-settings'),
    re_path(r'^myparams/(?P<my_params>[\w_-]+)/$', views.sample_view, name='sample-params'),
    re_path(_(r'^account/$'), views.sample_view, {'message': 'sample account page'}, name='sample-account'),
    re_path(r'^account/my_profile/$', views.sample_view, {'message': 'sample my profile page'}, name='sample-profile'),
    re_path(r'^category/(?P<id>[0-9]+)/$', views.category_view, name='category_view'),
    re_path(r'^notfound/$', views.notfound, name='notfound'),
    re_path(r'^extra_1/$', views.extra_view, {'message': 'test urlconf'}, name='extra_first'),
    re_path(r'^class-view/$', views.ClassView(), name='sample-class-view'),
    re_path(r'^class-based-view/$', views.ClassBasedView.as_view(), name='sample-class-based-view'),
    re_path(r'^', include('cms.test_utils.project.sampleapp.urls_extra'), {'opts': 'someopts'}),
]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0014_auto_20160404_1908'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cmsplugin',
            name='position',
            field=models.PositiveSmallIntegerField(default=0, verbose_name='position', editable=False),
        ),
    ]



from functools import partial
from logging import getLogger

from django.contrib import messages
from django.contrib.sites.models import Site
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.urls import NoReverseMatch
from django.utils.functional import cached_property
from django.utils.module_loading import autodiscover_modules
from django.utils.translation import (
    get_language_from_request,
)
from django.utils.translation import (
    gettext_lazy as _,
)

from cms.utils import get_current_site
from cms.utils.conf import get_cms_setting
from cms.utils.i18n import (
    get_default_language_for_site,
    is_language_prefix_patterns_used,
)
from menus.base import Menu
from menus.exceptions import NamespaceAlreadyRegistered
from menus.models import CacheKey

logger = getLogger('menus')


def _build_nodes_inner_for_one_menu(nodes, menu_class_name):
    '''
    This is an easier to test "inner loop" building the menu tree structure
    for one menu (one language, one site)
    '''
    done_nodes = {}  # Dict of node.id:Node
    final_nodes = []

    # This is to prevent infinite loops - we need to compare the number of
    # times we see a specific node to "something", and for the time being,
    # it's the total number of nodes
    list_total_length = len(nodes)

    while nodes:
        # For when the node has a parent_id but we haven't seen it yet.
        # We must not append it to the final list in this case!
        should_add_to_final_list = True

        node = nodes.pop(0)

        # Increment the "seen" counter for this specific node.
        node._counter = getattr(node, '_counter', 0) + 1

        # Implicit namespacing by menu.__name__
        if not node.namespace:
            node.namespace = menu_class_name
        if node.namespace not in done_nodes:
            # We need to create the namespace dict to avoid KeyErrors
            done_nodes[node.namespace] = {}

        # If we have seen the parent_id already...
        if node.parent_id in done_nodes[node.namespace]:
            # Implicit parent namespace by menu.__name__
            if not node.parent_namespace:
                node.parent_namespace = menu_class_name
            parent = done_nodes[node.namespace][node.parent_id]
            parent.children.append(node)
            node.parent = parent
        # If it has a parent_id but we haven't seen it yet...
        elif node.parent_id:
            # We check for infinite loops here, by comparing the number of
            # times we "saw" this node to the number of nodes in the list
            if node._counter < list_total_length:
                nodes.append(node)
            # Never add this node to the final list until it has a real
            # parent (node.parent)
            should_add_to_final_list = False

        if should_add_to_final_list:
            final_nodes.append(node)
            # add it to the "seen" list
            done_nodes[node.namespace][node.id] = node
    return final_nodes


def _get_menu_class_for_instance(menu_class, instance):
    """
    Returns a new menu class that subclasses
    menu_class but is bound to instance.
    This means it sets the "instance" attribute of the class.
    """
    attrs = {'instance': instance}
    class_name = menu_class.__name__
    meta_class = type(menu_class)
    return meta_class(class_name, (menu_class,), attrs)


class MenuRenderer(object):
    # The main logic behind this class is to decouple
    # the singleton menu pool from the menu rendering logic.
    # By doing this we can be sure that each request has its
    # private instance that will always have the same attributes.

    def __init__(self, pool, request):
        self.pool = pool
        # It's important this happens on init
        # because we need to make sure that a menu renderer
        # points to the same registered menus as long as the
        # instance lives.
        self.menus = pool.get_registered_menus(for_rendering=True)
        self.request = request
        self.request_language = None
        if is_language_prefix_patterns_used():
            self.request_language = get_language_from_request(request, check_path=True)
        if not self.request_language:
            self.request_language = get_default_language_for_site(get_current_site().pk)
        self.site = Site.objects.get_current(request)
        toolbar = getattr(request, "toolbar", None)
        self.edit_or_preview = toolbar.edit_mode_active or toolbar.preview_mode_active if toolbar else False

    @property
    def cache_key(self):
        prefix = get_cms_setting('CACHE_PREFIX')

        key = '%smenu_nodes_%s_%s' % (prefix, self.request_language, self.site.pk)

        if self.request.user.is_authenticated:
            key += '_%s_user' % self.request.user.pk

        if self.edit_or_preview:
            key += ':edit'
        else:
            key += ':public'
        return key

    @cached_property
    def is_cached(self):
        db_cache_key_lookup = CacheKey.objects.filter(
            key=self.cache_key,
            language=self.request_language,
            site=self.site.pk,
        )
        return db_cache_key_lookup.exists()

    def _build_nodes(self):
        """
        This is slow. Caching must be used.
        One menu is built per language and per site.

        Namespaces: they are ID prefixes to avoid node ID clashes when plugging
        multiple trees together.

        - We iterate on the list of nodes.
        - We store encountered nodes in a dict (with namespaces):
            done_nodes[<namespace>][<node's id>] = node
        - When a node has a parent defined, we lookup that parent in done_nodes
            if it's found:
                set the node as the node's parent's child (re-read this)
            else:
                the node is put at the bottom of the list
        """
        key = self.cache_key

        cached_nodes = cache.get(key, None)

        if cached_nodes and self.is_cached:
            # Only use the cache if the key is present in the database.
            # This prevents a condition where keys which have been removed
            # from the database due to a change in content, are still used.
            return cached_nodes

        final_nodes = []
        toolbar = getattr(self.request, 'toolbar', None)

        for menu_class_name in self.menus:
            menu = self.get_menu(menu_class_name)

            try:
                nodes = menu.get_nodes(self.request)
            except NoReverseMatch:
                # Apps might raise NoReverseMatch if an apphook does not yet
                # exist, skip them instead of crashing
                nodes = []
                if toolbar and toolbar.is_staff:
                    messages.error(
                        self.request,
                        _('Menu %s cannot be loaded. Please, make sure all its urls exist and can be resolved.') %
                        menu_class_name
                    )
                logger.error("Menu %s could not be loaded." % menu_class_name, exc_info=True)
            # nodes is a list of navigation nodes (page tree in cms + others)
            final_nodes += _build_nodes_inner_for_one_menu(nodes, menu_class_name)

        cache.set(key, final_nodes, get_cms_setting('CACHE_DURATIONS')['menus'])

        if not self.is_cached:
            # No need to invalidate the internal lookup cache,
            # just set the value directly.
            self.__dict__['is_cached'] = True
            # We need to have a list of the cache keys for languages and sites that
            # span several processes - so we follow the Django way and share through
            # the database. It's still cheaper than recomputing every time!
            # This way we can selectively invalidate per-site and per-language,
            # since the cache is shared but the keys aren't
            CacheKey.objects.create(key=key, language=self.request_language, site=self.site.pk)
        return final_nodes

    def _mark_selected(self, nodes):
        for node in nodes:
            node.selected = node.is_selected(self.request)
        return nodes

    def apply_modifiers(self, nodes, namespace=None, root_id=None, post_cut=False, breadcrumb=False):
        if not post_cut:
            nodes = self._mark_selected(nodes)

        # Only fetch modifiers when they're needed.
        # We can do this because unlike menu classes,
        # modifiers can't change on a request basis.
        for cls in self.pool.get_registered_modifiers():
            inst = cls(renderer=self)
            nodes = inst.modify(
                self.request, nodes, namespace, root_id, post_cut, breadcrumb)
        return nodes

    def get_nodes(self, namespace=None, root_id=None, breadcrumb=False):
        nodes = self._build_nodes()
        nodes = self.apply_modifiers(
            nodes=nodes,
            namespace=namespace,
            root_id=root_id,
            post_cut=False,
            breadcrumb=breadcrumb,
        )
        return nodes

    def get_menu(self, menu_name):
        MenuClass = self.menus[menu_name]
        return MenuClass(renderer=self)


class MenuPool(object):

    def __init__(self):
        self.menus = {}
        self.modifiers = []
        self.discovered = False

    def get_renderer(self, request):
        self.discover_menus()
        # Returns a menu pool wrapper that is bound
        # to the given request and can perform
        # operations based on the given request.
        return MenuRenderer(pool=self, request=request)

    def discover_menus(self):
        if self.discovered:
            return
        autodiscover_modules('cms_menus')
        from menus.modifiers import register
        register()
        self.discovered = True

    def get_registered_menus(self, for_rendering=False):
        """
        Returns all registered menu classes.

        :param for_rendering: Flag that when True forces us to include
            all CMSAttachMenu subclasses, even if they're not attached.
        """
        self.discover_menus()
        registered_menus = {}

        for menu_class_name, menu_cls in self.menus.items():
            if isinstance(menu_cls, Menu):
                # A Menu **instance** was registered,
                # this is non-standard, but acceptable.
                menu_cls = menu_cls.__class__
            if hasattr(menu_cls, "get_instances"):
                # It quacks like a CMSAttachMenu.
                # Expand the one CMSAttachMenu into multiple classes.
                # Each class is bound to the instance the menu is attached to.
                _get_menu_class = partial(_get_menu_class_for_instance, menu_cls)

                instances = menu_cls.get_instances() or []
                for instance in instances:
                    # For each instance, we create a unique class
                    # that is bound to that instance.
                    # Doing this allows us to delay the instantiation
                    # of the menu class until it's needed.
                    # Plus we keep the menus consistent by always
                    # pointing to a class instead of an instance.
                    namespace = "{0}:{1}".format(
                        menu_class_name, instance.pk)
                    registered_menus[namespace] = _get_menu_class(instance)

                if not instances and not for_rendering:
                    # The menu is a CMSAttachMenu but has no instances,
                    # normally we'd just ignore it but it's been
                    # explicitly set that we are not rendering these menus
                    # via the (for_rendering) flag.
                    registered_menus[menu_class_name] = menu_cls
            elif hasattr(menu_cls, "get_nodes"):
                # This is another type of Menu, cannot be expanded, but must be
                # instantiated, none-the-less.
                registered_menus[menu_class_name] = menu_cls
            else:
                raise ValidationError(
                    "Something was registered as a menu, but isn't.")
        return registered_menus

    def get_registered_modifiers(self):
        return self.modifiers

    def clear(self, site_id=None, language=None, all=False):
        '''
        This invalidates the cache for a given menu (site_id and language)
        '''
        if all:
            cache_keys = CacheKey.objects.get_keys()
        else:
            cache_keys = CacheKey.objects.get_keys(site_id, language)

        to_be_deleted = cache_keys.distinct().values_list('key', flat=True)

        if to_be_deleted:
            cache.delete_many(to_be_deleted)
            cache_keys.delete()

    def register_menu(self, menu_cls):
        from menus.base import Menu
        assert issubclass(menu_cls, Menu)
        if menu_cls.__name__ in self.menus:
            raise NamespaceAlreadyRegistered(
                "[{0}] a menu with this name is already registered".format(
                    menu_cls.__name__))
        # Note: menu_cls should still be the menu CLASS at this point.
        self.menus[menu_cls.__name__] = menu_cls

    def register_modifier(self, modifier_class):
        from menus.base import Modifier
        assert issubclass(modifier_class, Modifier)
        if modifier_class not in self.modifiers:
            self.modifiers.append(modifier_class)

    def get_menus_by_attribute(self, name, value):
        """
        Returns the list of menus that match the name/value criteria provided.
        """
        # Note that we are limiting the output to only single instances of any
        # specific menu class. This is to address issue (#4041) which has
        # cropped-up in 3.0.13/3.0.0.
        # By setting for_rendering to False
        # we're limiting the output to menus
        # that are registered and have instances
        # (in case of attached menus).
        menus = self.get_registered_menus(for_rendering=False)
        return sorted(list(set([(menu.__name__, menu.name)
                                for menu_class_name, menu in menus.items()
                                if getattr(menu, name, None) == value])))

    def get_nodes_by_attribute(self, nodes, name, value):
        return [node for node in nodes if node.attr.get(name, None) == value]


menu_pool = MenuPool()



from django.utils.encoding import smart_str


class Menu(object):
    """The base class for all menu-generating classes."""
    namespace = None

    def __init__(self, renderer):
        self.renderer = renderer

        if not self.namespace:
            self.namespace = self.__class__.__name__

    def get_nodes(self, request):
        """Each subclass of Menu should return a list of :class:`menus.base.NavigationNode` instances."""
        raise NotImplementedError


class Modifier(object):
    """The base class for all menu-modifying classes. A modifier add, removes or changes
    :class:`menus.base.NavigationNode` in the list."""
    def __init__(self, renderer):
        self.renderer = renderer

    def modify(self, request, nodes, namespace, root_id, post_cut, breadcrumb):
        """Each subclass of :class:`Modifier` should implement a :meth:`modify` method."""
        pass


class NavigationNode(object):
    """
    Each node in a menu tree is represented by a ``NavigationNode`` instance.
    """
    selected = None
    sibling = False
    ancestor = False
    descendant = False

    def __init__(self, title, url, id, parent_id=None, parent_namespace=None,
                 attr=None, visible=True):
        """
        :param str title: The title to display this menu item with.
        :param str url: The URL associated with this menu item.
        :param id: Unique (for the current tree) ID of this item.
        :param parent_id: Optional, ID of the parent item.
        :param parent_namespace: Optional, namespace of the parent.
        :param dict attr: Optional, dictionary of additional information to store on
                          this node.
        :param bool visible: Optional, defaults to ``True``, whether this item is
                             visible or not.
        """
        self.children = []  # do not touch
        self.parent = None  # do not touch, code depends on this
        self.namespace = None  # TODO: Assert why we need this and above
        self.title = title
        self.url = url
        self.id = id
        self.parent_id = parent_id
        self.parent_namespace = parent_namespace
        self.visible = visible
        self.attr = attr or {}  # To avoid declaring a dict in defaults...
        """
        A dictionary, provided in order that arbitrary attributes may be added to the node -
        placing them directly on the node itself could cause a clash with an existing or future attribute.

        An important key in this dictionary is ``is_page``: if ``True``, the node represents a django CMS
        ``Page`` object.

        Nodes that represent CMS pages have the following keys in ``attr``:

        * **auth_required** (*bool*) – is authentication required to access this page
        * **is_page** (*bool*) – Always True
        * **navigation_extenders** (*list*) – navigation extenders connected to this node
        * **redirect_url** (*str*) – redirect URL of page (if any)
        * **reverse_id** (*str*) – unique identifier for the page
        * **soft_root** (*bool*) – whether page is a soft root
        * **visible_for_authenticated** (*bool*) – visible for authenticated users
        * **visible_for_anonymous** (*bool*) – visible for anonymous users
        """

    def __repr__(self):
        return "<Navigation Node: %s>" % smart_str(self.title)

    def get_menu_title(self):
        """Utility method to return the associated title, using the same naming
        convention used by :class:`cms.models.pagemodel.Page`."""
        return self.title

    def get_absolute_url(self):
        """Utility method to return the URL associated with this menu item,
        primarily to follow naming convention asserted by Django."""
        return self.url

    def get_attribute(self, name):
        """Retrieves a dictionary item from :attr:`attr`. Returns ``None``  if it does not exist."""
        return self.attr.get(name, None)

    def get_descendants(self):
        """Returns a list of all children beneath the current menu item."""
        return sum(([node] + node.get_descendants() for node in self.children), [])

    def get_ancestors(self):
        """Returns a list of all parent items, excluding the current menu item."""
        if getattr(self, 'parent', None):
            return [self.parent] + self.parent.get_ancestors()
        else:
            return []

    def is_selected(self, request):
        node_abs_url = self.get_absolute_url()
        return node_abs_url == request.path



from django.db import models

from cms.models import CMSPlugin


class Article(models.Model):
    title = models.CharField(max_length=50)
    section = models.ForeignKey('Section', on_delete=models.CASCADE)

    def __str__(self):
        return u"%s -- %s" % (self.title, self.section)


class Section(models.Model):
    name = models.CharField(max_length=50)

    def __str__(self):
        return self.name


class ArticlePluginModel(CMSPlugin):
    title = models.CharField(max_length=50)
    sections = models.ManyToManyField('Section')

    def __str__(self):
        return self.title

    def copy_relations(self, oldinstance):
        self.sections.set(oldinstance.sections.all())


class FKModel(models.Model):
    fk_field = models.ForeignKey('PluginModelWithFKFromModel', on_delete=models.CASCADE)


class M2MTargetModel(models.Model):
    title = models.CharField(max_length=50)


class PluginModelWithFKFromModel(CMSPlugin):
    title = models.CharField(max_length=50)

    def copy_relations(self, oldinstance):
        # Like suggested in the docs
        for associated_item in oldinstance.fkmodel_set.all():
            associated_item.pk = None
            associated_item.fk_field = self
            associated_item.save()


class PluginModelWithM2MToModel(CMSPlugin):
    m2m_field = models.ManyToManyField(M2MTargetModel)

    def copy_relations(self, oldinstance):
        # Like suggested in the docs
        self.m2m_field.set(oldinstance.m2m_field.all())



from django.conf import settings
from django.db import models

from cms.extensions import PageContentExtension, PageExtension
from cms.extensions.extension_pool import extension_pool


class MyPageExtension(PageExtension):
    extra = models.CharField(blank=True, default='', max_length=255)
    favorite_users = models.ManyToManyField(settings.AUTH_USER_MODEL, blank=True)

    def copy_relations(self, other, language):
        for favorite_user in other.favorite_users.all():
            favorite_user.pk = None
            favorite_user.mypageextension = self
            favorite_user.save()


extension_pool.register(MyPageExtension)


# extension_pool.register can be used also as a decorator
@extension_pool.register
class MyPageContentExtension(PageContentExtension):
    extra_title = models.CharField(blank=True, default='', max_length=255)


class MultiTablePageExtensionParent(models.Model):
    extension_parent_field = models.CharField(blank=True, default='', max_length=255)


class MultiTablePageExtension(MultiTablePageExtensionParent, PageExtension):
    multitable_extra = models.CharField(blank=True, default='', max_length=255)


extension_pool.register(MultiTablePageExtension)


class MultiTableContentExtensionParent(models.Model):
    extension_content_parent_field = models.CharField(blank=True, default='', max_length=255)


class MultiTablePageContentExtension(MultiTableContentExtensionParent, PageContentExtension):
    multitable_extra_content = models.CharField(blank=True, default='', max_length=255)


extension_pool.register(MultiTablePageContentExtension)



# Generated by Django 1.11.12 on 2018-05-29 14:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0024_set_plugin_absolute_position'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='cmsplugin',
            options={'ordering': ('position',)},
        ),
        migrations.AlterField(
            model_name='cmsplugin',
            name='position',
            field=models.SmallIntegerField(default=1, editable=False, verbose_name='position'),
        ),
        migrations.AlterUniqueTogether(
            name='cmsplugin',
            unique_together=set([('placeholder', 'language', 'position')]),
        ),
        migrations.AddIndex(
            model_name='cmsplugin',
            index=models.Index(fields=['placeholder', 'language', 'position'], name='cms_cmsplug_placeho_3f2b0f_idx'),
        ),
        migrations.RemoveField(
            model_name='cmsplugin',
            name='depth',
        ),
        migrations.RemoveField(
            model_name='cmsplugin',
            name='numchild',
        ),
        migrations.RemoveField(
            model_name='cmsplugin',
            name='path',
        ),
    ]



import inspect
from contextlib import contextmanager
from itertools import chain

from django.conf import settings
from django.utils.decorators import method_decorator
from django.utils.termcolors import colorize
from sekizai.helpers import validate_template

from cms import constants
from cms.models import AliasPluginModel
from cms.utils.compat.dj import is_installed
from cms.utils.conf import get_cms_setting

SUCCESS = 1
WARNING = 2
ERROR = 3
SKIPPED = 4

CHECKERS = []


class FileOutputWrapper:
    """
    Wraps two file-like objects (that support at the very least the 'write'
    method) into an API to be used by the check function further down in
    this module.

    The following properties are public (and required) by alternative implementations:

        errors: integer count of errors encountered
        successes: integer count of successes encountered
        warnings: integer count of warnings encountered
        skips: integer count of skips encountered
        successful: Whether the checks were successful (no errors)

    They must also provide these methods:

        write_line(message=''): writes a message to stdout
        write_stderr_line(message=''): writes a message to stderr
        success(message): reports and registers a successful check
        error(message): reports and registers an error
        warn(message); reports and registers a warning
        skip(message): reports and registers a skipped check
        section(title): A context manager that starts a new section. For the
            Section API see FileSectionWrapper
    """
    def __init__(self, stdout, stderr):
        self.stdout = stdout
        self.stderr = stderr
        self.section_wrapper = FileSectionWrapper
        self.errors = 0
        self.successes = 0
        self.warnings = 0
        self.skips = 0

    def colorize(self, msg, opts=(), **kwargs):
        return colorize(msg, opts=opts, **kwargs)

    def write_line(self, message=''):
        self.write(u'%s\n' % message)

    def write(self, message):
        self.stdout.write(message)

    def write_stderr_line(self, message=''):
        self.write_stderr(u'%s\n' % message)

    def write_stderr(self, message):
        self.stderr.write(message)

    def success(self, message):
        self.successes += 1
        self.write_line(u'%s %s' % (message, self.colorize('[OK]', fg='green', opts=['bold'])))

    def error(self, message):
        self.errors += 1
        self.write_stderr_line(u'%s %s' % (message, self.colorize('[ERROR]', fg='red', opts=['bold'])))

    def warn(self, message):
        self.warnings += 1
        self.write_stderr_line(u'%s %s' % (message, self.colorize('[WARNING]', fg='yellow', opts=['bold'])))

    def skip(self, message):
        self.skips += 1
        self.write_line(u'%s %s' % (message, self.colorize('[SKIP]', fg='blue', opts=['bold'])))

    @method_decorator(contextmanager)
    def section(self, title):
        self.write_line(self.colorize(title, opts=['bold']))
        self.write_line(self.colorize('=' * len(title), opts=['bold']))
        self.write_line()
        wrapper = self.section_wrapper(self)
        try:
            yield wrapper
        except:  # NOQA
            self.error('Checker failed, see traceback')
            raise
        self.errors += wrapper.errors
        self.successes += wrapper.successes
        self.warnings += wrapper.warnings
        self.skips += wrapper.skips
        self.write_line('')

    @property
    def successful(self):
        return not self.errors


class FileSectionWrapper(FileOutputWrapper):
    """
    Used from FileOutputWrapper to report checks in a section.

    If you want to provide your own output class, you may want to subclass
    this class for the section reporting too. If you want to use your own,
    you must defined at least the same API as FileOutputWrapper, as well
    as these four additional methods:

        finish_success(message): End the section (successfully)
        finish_error(message): End the section with errors
        finish_warning(message): End this section with a warning
        finish_skip(message): End this (skipped) section
    """
    def __init__(self, wrapper):
        super().__init__(wrapper.stdout, wrapper.stderr)
        self.wrapper = wrapper

    def write_line(self, message=''):
        self.write(u'  - %s\n' % message)

    def write_stderr_line(self, message=''):
        self.write_stderr(u'  - %s\n' % message)

    def finish_success(self, message):
        self.wrapper.write_line()
        self.wrapper.success(message)

    def finish_error(self, message):
        self.wrapper.write_line()
        self.wrapper.error(message)

    def finish_warning(self, message):
        self.wrapper.write_line()
        self.wrapper.warning(message)

    def finish_skip(self, message):
        self.wrapper.write_lin()
        self.wrapper.skip(message)


def define_check(func):
    """
    Helper decorator to register a check function.
    """
    CHECKERS.append(func)
    return func


@define_check
def check_sekizai(output):
    with output.section("Sekizai") as section:
        sekizai_installed = is_installed('sekizai')

        if sekizai_installed:
            section.success("Sekizai is installed")
        else:
            section.error("Sekizai is not installed, could not find 'sekizai' in INSTALLED_APPS")
        processors = list(
            chain(*[template['OPTIONS'].get('context_processors', []) for template in settings.TEMPLATES]))
        if 'sekizai.context_processors.sekizai' in processors:
            section.success("Sekizai template context processor is installed")
        else:
            section.error("Sekizai template context processor is not installed, could not find "
                          "'sekizai.context_processors.sekizai' in TEMPLATES option context_processors")

        if not sekizai_installed:
            # sekizai is not installed.
            # we can't reliable check templates
            # because template loading won't work
            return

        for template, _ in get_cms_setting('TEMPLATES'):
            if template == constants.TEMPLATE_INHERITANCE_MAGIC:
                continue
            if validate_template(template, ['js', 'css']):
                section.success("Sekizai namespaces 'js' and 'css' found in %r" % template)
            else:
                section.error("Sekizai namespaces 'js' and 'css' not found in %r" % template)
        if section.successful:
            section.finish_success("Sekizai configuration okay")
        else:
            section.finish_error("Sekizai configuration has errors")


@define_check
def check_i18n(output):
    with output.section("Internationalization") as section:
        if isinstance(getattr(settings, 'CMS_LANGUAGES', {}), dict):
            section.success("New style CMS_LANGUAGES")
        else:
            section.warn("Old style (tuple based) CMS_LANGUAGES, please switch to the new (dictionary based) style")
        if getattr(settings, 'LANGUAGE_CODE', '').find('_') > -1:
            section.warn("LANGUAGE_CODE must contain a valid language code, not a locale (e.g.: 'en-us' instead of "
                         "'en_US'): '%s' provided" % getattr(settings, 'LANGUAGE_CODE', ''))
        for lang in getattr(settings, 'LANGUAGES', ()):
            if lang[0].find('_') > -1:
                section.warn("LANGUAGES must contain valid language codes, not locales (e.g.: 'en-us' instead of "
                             "'en_US'): '%s' provided" % lang[0])
        if settings.SITE_ID == hash(settings.SITE_ID):
            for site, items in get_cms_setting('LANGUAGES').items():
                if type(site) == int:
                    for lang in items:
                        if lang['code'].find('_') > -1:
                            section.warn("CMS_LANGUAGES entries must contain valid language codes, not locales (e.g.: "
                                         "'en-us' instead of 'en_US'): '%s' provided" % lang['code'])
        else:
            section.error("SITE_ID must be an integer, not %r" % settings.SITE_ID)


@define_check
def check_middlewares(output):
    with output.section("Middlewares") as section:
        required_middlewares = (
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.middleware.csrf.CsrfViewMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.contrib.messages.middleware.MessageMiddleware',
            'django.middleware.locale.LocaleMiddleware',
            'django.middleware.common.CommonMiddleware',
            'cms.middleware.user.CurrentUserMiddleware',
            'cms.middleware.page.CurrentPageMiddleware',
            'cms.middleware.toolbar.ToolbarMiddleware',
            'cms.middleware.language.LanguageCookieMiddleware',
        )
        if getattr(settings, 'MIDDLEWARE', None):
            middlewares = settings.MIDDLEWARE
        else:
            middlewares = settings.MIDDLEWARE_CLASSES
        for middleware in required_middlewares:
            if middleware not in middlewares:
                section.error("%s middleware must be in MIDDLEWARE_CLASSES" % middleware)


@define_check
def check_context_processors(output):
    with output.section("Context processors") as section:
        processors = list(
            chain(*[template['OPTIONS'].get('context_processors', []) for template in settings.TEMPLATES]))
        required_processors = (
            'cms.context_processors.cms_settings',
        )
        for processor in required_processors:
            if processor not in processors:
                section.error("%s context processor must be in TEMPLATES option context_processors" % processor)


@define_check
def check_plugin_instances(output):
    from cms.management.commands.subcommands.list import plugin_report
    with output.section("Plugin instances") as section:
        # get the report
        report = plugin_report()
        section.success("Plugin instances of %s types found in the database" % len(report))
        # loop over plugin types in the report
        for plugin_type in report:
            # warn about those that are not installed
            if not plugin_type["model"]:
                section.error("%s has instances but is no longer installed" % plugin_type["type"])
            # warn about those that have unsaved instances
            if plugin_type["unsaved_instances"]:
                section.error(
                    "%s has %s unsaved instances" % (plugin_type["type"], len(plugin_type["unsaved_instances"])))

        if section.successful:
            section.finish_success("The plugins in your database are in good order")
        else:
            section.finish_error("There are potentially serious problems with the plugins in your database. \nEven if "
                                 "your site works, you should run the 'manage.py cms list plugins' \ncommand and then "
                                 "the 'manage.py cms delete-orphaned-plugins' command. \nThis will alter your "
                                 "database; read the documentation before using it.")


@define_check
def check_copy_relations(output):
    from cms.extensions import extension_pool
    from cms.extensions.models import BaseExtension
    from cms.models.pluginmodel import CMSPlugin
    from cms.plugin_pool import plugin_pool

    def c_to_s(klass):
        return '%s.%s' % (klass.__module__, klass.__name__)

    def get_class(method_name, model):
        for cls in inspect.getmro(model):
            if method_name in cls.__dict__:
                return cls
        return None

    with output.section('Presence of "copy_relations"') as section:
        plugin_pool.discover_plugins()
        for plugin in plugin_pool.plugins.values():
            plugin_class = plugin.model
            if get_class('copy_relations', plugin_class) is not CMSPlugin or plugin_class is CMSPlugin:
                # this class defines a ``copy_relations`` method, nothing more
                # to do
                continue
            for rel in plugin_class._meta.many_to_many:
                section.warn('%s has a many-to-many relation to %s,\n    but no "copy_relations" method defined.' % (
                    c_to_s(plugin_class),
                    c_to_s(rel.model),
                ))
            for rel in plugin_class._get_related_objects():
                if rel.model != CMSPlugin and not issubclass(
                        rel.model, plugin.model) and rel.model != AliasPluginModel:
                    section.warn('%s has a foreign key from %s,\n    but no "copy_relations" method defined.' % (
                        c_to_s(plugin_class),
                        c_to_s(rel.model),
                    ))

        for extension in chain(extension_pool.page_extensions, extension_pool.page_content_extensions):
            if get_class('copy_relations', extension) is not BaseExtension:
                # OK, looks like there is a 'copy_relations' defined in the
                # extension... move along...
                continue
            for rel in extension._meta.many_to_many:
                section.warn(
                    '%s has a many-to-many relation to %s,\n    '
                    'but no "copy_relations" method defined.' % (
                        c_to_s(extension),
                        c_to_s(rel.remote_field.model),
                    )
                )
            for rel in extension._get_related_objects():
                if rel.model != extension:
                    section.warn('%s has a foreign key from %s,\n    but no "copy_relations" method defined.' % (
                        c_to_s(extension),
                        c_to_s(rel.model),
                    ))

        if not section.warnings:
            section.finish_success('All plugins and page/title extensions have "copy_relations" method if needed.')
        else:
            section.finish_success('Some plugins or page/title extensions do not define a "copy_relations" method.\n'
                                   'This might lead to data loss when publishing or copying plugins/extensions.\n'
                                   'See https://django-cms.readthedocs.io/en/latest/extending_cms/custom_plugins.html#handling-relations or '  # noqa
                                   'https://django-cms.readthedocs.io/en/latest/extending_cms/extending_page_title.html#handling-relations.')  # noqa


def check(output):
    """
    Checks the configuration/environment of this django CMS installation.

    'output' should be an object that provides the same API as FileOutputWrapper.

    Returns whether the configuration/environment are okay (has no errors)
    """
    title = "Checking django CMS installation"
    border = '*' * len(title)
    output.write_line(output.colorize(border, opts=['bold']))
    output.write_line(output.colorize(title, opts=['bold']))
    output.write_line(output.colorize(border, opts=['bold']))
    output.write_line()
    for checker in CHECKERS:
        checker(output)
    output.write_line()
    with output.section("OVERALL RESULTS"):
        if output.errors:
            output.write_stderr_line(output.colorize("%s errors!" % output.errors, opts=['bold'], fg='red'))
        if output.warnings:
            output.write_stderr_line(output.colorize("%s warnings!" % output.warnings, opts=['bold'], fg='yellow'))
        if output.skips:
            output.write_line(output.colorize("%s checks skipped!" % output.skips, opts=['bold'], fg='blue'))
        output.write_line(output.colorize("%s checks successful!" % output.successes, opts=['bold'], fg='green'))
        output.write_line()
        if output.errors:
            output.write_stderr_line(output.colorize('Please check the errors above', opts=['bold'], fg='red'))
        elif output.warnings:
            output.write_stderr_line(output.colorize('Installation okay, but please check warnings above',
                                                     opts=['bold'], fg='yellow'))
        else:
            output.write_line(output.colorize('Installation okay', opts=['bold'], fg='green'))
    return output.successful
